"""Add 'table_name' column

Revision ID: f4fd4c80f40a
Revises: d82e1d06a82c
Create Date: 2019-05-28 11:34:52.353749

"""
from alembic import op
import sqlalchemy as sa


revision = 'f4fd4c80f40a'
down_revision = 'd82e1d06a82c'
branch_labels = None
depends_on = None


def upgrade():
    url = op.get_bind().engine.url
    collation = 'utf8_bin' if 'mysql' in url.drivername else None
    op.add_column(
        'commcare_export_runs',
        sa.Column('table_name', sa.Unicode(255, collation=collation))
    )

def downgrade():
    op.drop_column('commcare_export_runs', 'table_name')



"""Add pagination_mode to checkpoint

Revision ID: 6f158d161ab6
Revises: a56c82a8d02e
Create Date: 2021-01-25 15:13:45.996453

"""
from alembic import op
import sqlalchemy as sa


revision = '6f158d161ab6'
down_revision = 'a56c82a8d02e'
branch_labels = None
depends_on = None



def upgrade():
    url = op.get_bind().engine.url
    collation = 'utf8_bin' if 'mysql' in url.drivername else None
    op.add_column(
        'commcare_export_runs',
        sa.Column('pagination_mode', sa.Unicode(255, collation=collation))
    )

def downgrade():
    op.drop_column('commcare_export_runs', 'pagination_mode')



import io
import re
import os.path
import subprocess

__all__ = ['__version__', 'stored_version', 'git_version']

VERSION_PATH = os.path.join(os.path.dirname(__file__), 'VERSION')


def stored_version():
    if os.path.exists(VERSION_PATH):
        with io.open(VERSION_PATH, encoding='ascii') as fh:
            return fh.read().strip()
    else:
        return None


def git_version():
    described_version_bytes = subprocess.Popen(
        ['git', 'describe'],
        stdout=subprocess.PIPE
    ).communicate()[0].strip()
    version_raw = described_version_bytes.decode('ascii')
    return parse_version(version_raw)


def parse_version(version_raw):
    """Attempt to convert a git version to a version
    compatible with PEP440: https://peps.python.org/pep-0440/
    """
    match = re.match('(\d+\.\d+\.\d+)(?:-(\d+).*)?', version_raw)
    if match:
        tag_version, lead_count = match.groups()
        if lead_count:
            tag_version += ".dev{}".format(lead_count)
        return tag_version

    return version_raw


def version():
    return stored_version() or git_version()


__version__ = version()

if __name__ == '__main__':
    print(__version__)



import datetime
import logging
import os
import uuid
from contextlib import contextmanager
from operator import attrgetter

import dateutil.parser
from sqlalchemy import Boolean, Column, String, and_, func
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import sessionmaker

from commcare_export.commcare_minilinq import PaginationMode
from commcare_export.exceptions import DataExportException
from commcare_export.writers import SqlMixin

logger = logging.getLogger(__name__)
repo_root = os.path.abspath(os.path.join(__file__, os.pardir, os.pardir))

Base = declarative_base()


class Checkpoint(Base):
    __tablename__ = 'commcare_export_runs'

    id = Column(String, primary_key=True)
    query_file_name = Column(String)
    query_file_md5 = Column(String)
    table_name = Column(String)
    key = Column(String)
    project = Column(String)
    commcare = Column(String)
    since_param = Column(String)
    time_of_run = Column(String)
    final = Column(Boolean)
    data_source = Column(String)
    last_doc_id = Column(String)
    pagination_mode = Column(String)
    cursor = Column(String)

    def get_pagination_mode(self):
        """
        Get Enum from value stored in the checkpoint. Null or empty
        value defaults to 'date_modified' mode to support legacy
        checkpoints.
        """
        if not self.pagination_mode:
            return PaginationMode.date_modified

        return PaginationMode[self.pagination_mode]

    def __repr__(self):
        return (
            "<Checkpoint("
            "id={r.id}, "
            "query_file_name={r.query_file_name}, "
            "query_file_md5={r.query_file_md5}, "
            "table_name={r.table_name}, "
            "key={r.key}, "
            "project={r.project}, "
            "commcare={r.commcare}, "
            "since_param={r.since_param}, "
            "time_of_run={r.time_of_run}, "
            "final={r.final}), "
            "data_source={r.data_source}, "
            "last_doc_id={r.last_doc_id}, "
            "pagination_mode={r.pagination_mode},"
            "cursor={r.cursor}>"
        ).format(r=self)


@contextmanager
def session_scope(Session):
    """
    Provide a transactional scope around a series of operations.
    """
    session = Session()
    try:
        yield session
        session.commit()
    except:
        session.rollback()
        raise
    finally:
        session.close()


class CheckpointManager(SqlMixin):
    table_name = 'commcare_export_runs'
    migrations_repository = os.path.join(repo_root, 'migrations')

    def __init__(
        self,
        db_url,
        query,
        query_md5,
        project,
        commcare,
        key=None,
        table_names=None,
        poolclass=None,
        engine=None,
        data_source=None
    ):
        super(CheckpointManager, self).__init__(
            db_url, poolclass=poolclass, engine=engine
        )
        self.query = query
        self.query_md5 = query_md5
        self.project = project
        self.commcare = commcare
        self.key = key
        self.Session = sessionmaker(self.engine, expire_on_commit=False)
        self.table_names = table_names
        self.data_source = data_source

    def for_dataset(self, data_source, table_names):
        return CheckpointManager(
            self.db_url,
            self.query,
            self.query_md5,
            self.project,
            self.commcare,
            self.key,
            engine=self.engine,
            table_names=table_names,
            data_source=data_source
        )

    def set_checkpoint(
        self,
        checkpoint_time,
        pagination_mode,
        is_final=False,
        doc_id=None,
        cursor=None,
    ):
        self._set_checkpoint(
            checkpoint_time,
            pagination_mode,
            is_final,
            doc_id=doc_id,
            cursor=cursor,
        )
        if is_final:
            self._cleanup()

    def _set_checkpoint(
        self,
        checkpoint_time,
        pagination_mode,
        final,
        time_of_run=None,
        doc_id=None,
        cursor=None,
    ):
        logger.info(
            'Setting %s checkpoint: data_source: %s, tables: %s, '
            'pagination_mode: %s, checkpoint: %s:%s',
            #
            'final' if final else 'batch',
            self.data_source,
            ', '.join(self.table_names),
            pagination_mode.name,
            checkpoint_time,
            doc_id,
        )
        if not checkpoint_time:
            raise DataExportException(
                'Tried to set an empty checkpoint. This is not allowed.'
            )
        self._validate_tables()

        if isinstance(checkpoint_time, str):
            since_param = checkpoint_time
        else:
            since_param = checkpoint_time.isoformat()

        created = []
        with session_scope(self.Session) as session:
            for table in self.table_names:
                checkpoint = Checkpoint(
                    id=uuid.uuid4().hex,
                    query_file_name=self.query,
                    query_file_md5=self.query_md5,
                    table_name=table,
                    key=self.key,
                    project=self.project,
                    commcare=self.commcare,
                    since_param=since_param,
                    time_of_run=time_of_run
                    or datetime.datetime.utcnow().isoformat(),
                    final=final,
                    data_source=self.data_source,
                    last_doc_id=doc_id,
                    pagination_mode=pagination_mode.name,
                    cursor=cursor,
                )
                session.add(checkpoint)
                created.append(checkpoint)
        return created

    def create_checkpoint_table(self, revision='head'):
        from alembic import command, config
        cfg = config.Config(
            os.path.join(self.migrations_repository, 'alembic.ini')
        )
        cfg.set_main_option('script_location', self.migrations_repository)
        with self.engine.begin() as connection:
            cfg.attributes['connection'] = connection
            command.upgrade(cfg, revision)

    def _cleanup(self):
        self._validate_tables()
        with session_scope(self.Session) as session:
            session.query(Checkpoint).filter_by(
                final=False,
                query_file_md5=self.query_md5,
                project=self.project,
                commcare=self.commcare
            ).filter(Checkpoint.table_name.in_(self.table_names)
                    ).delete(synchronize_session='fetch')

    def get_time_of_last_checkpoint(self, log_warnings=True):
        """
        Return the earliest time from the list of checkpoints that for
        the current query file / key.
        """
        run = self.get_last_checkpoint()
        if run and log_warnings:
            self.log_warnings(run)
        return run.since_param if run else None

    def get_last_checkpoint(self):
        """
        Return a single checkpoint such that it has the earliest
        `since_param` of all checkpoints for the active tables.
        """
        self._validate_tables()
        table_runs = []
        with session_scope(self.Session) as session:
            for table in self.table_names:
                if self.key:
                    table_run = self._get_last_checkpoint(
                        session,
                        table_name=table,
                        key=self.key,
                        project=self.project,
                        commcare=self.commcare
                    )
                else:
                    table_run = self._get_last_checkpoint(
                        session,
                        table_name=table,
                        query_file_md5=self.query_md5,
                        project=self.project,
                        commcare=self.commcare,
                        key=self.key
                    )
                if table_run:
                    table_runs.append(table_run)

        if not table_runs:
            table_runs = self.get_legacy_checkpoints()

        if table_runs:
            sorted_runs = list(
                sorted(table_runs, key=attrgetter('time_of_run'))
            )
            return sorted_runs[0]

    def get_legacy_checkpoints(self):
        with session_scope(self.Session) as session:
            # check without table_name
            table_run = self._get_last_checkpoint(
                session,
                query_file_md5=self.query_md5,
                table_name=None,
                project=self.project,
                commcare=self.commcare,
                key=self.key
            )
            if table_run:
                return self._set_checkpoint(
                    table_run.since_param,
                    PaginationMode.date_modified,
                    table_run.final,
                    table_run.time_of_run
                )

            # Check for run without the args
            table_run = self._get_last_checkpoint(
                session,
                query_file_md5=self.query_md5,
                key=self.key,
                project=None,
                commcare=None,
                table_name=None
            )
            if table_run:
                return self._set_checkpoint(
                    table_run.since_param,
                    PaginationMode.date_modified,
                    table_run.final,
                    table_run.time_of_run
                )

    def _get_last_checkpoint(self, session, **kwarg_filters):
        query = session.query(Checkpoint)
        if kwarg_filters:
            query = query.filter_by(**kwarg_filters)
        return query.order_by(Checkpoint.time_of_run.desc()).first()

    def log_warnings(self, run: Checkpoint) -> None:
        md5_mismatch = run.query_file_md5 != self.query_md5
        name_mismatch = run.query_file_name != self.query
        if md5_mismatch or name_mismatch:
            logger.warning(
                "Query differs from most recent checkpoint:\n"
                "From checkpoint:         name=%s, md5=%s\n"
                "From command line args:  name=%s, md5=%s\n",
                #
                run.query_file_name,
                run.query_file_md5,
                self.query,
                self.query_md5
            )

    def list_checkpoints(self, limit=20):
        """
        List all checkpoints filtered by:
        * file name
        * project
        * commcare
        * key

        Don't filter by MD5 on purpose.
        """
        with session_scope(self.Session) as session:
            query = self._filter_query(session.query(Checkpoint))
            if self.query:
                query = query.filter(Checkpoint.query_file_name == self.query)
            return query.order_by(Checkpoint.time_of_run.desc())[:limit]

    def _filter_query(self, query):
        if self.project:
            query = query.filter(Checkpoint.project == self.project)
        if self.commcare:
            query = query.filter(Checkpoint.commcare == self.commcare)
        if self.key:
            query = query.filter(Checkpoint.key == self.key)
        return query

    def get_latest_checkpoints(self):
        """
        Returns the latest checkpoint for each table filtered by the
        fields set in the manager:
        * query_md5
        * project
        * commcare
        * key
        """
        with session_scope(self.Session) as session:
            cols = [
                Checkpoint.project,
                Checkpoint.commcare,
                Checkpoint.query_file_md5,
                Checkpoint.table_name
            ]
            inner_query = self._filter_query(
                session.query(
                    *(
                        cols + [
                            func.max(Checkpoint.time_of_run
                                    ).label('max_time_of_run')
                        ]
                    )
                ).filter(Checkpoint.query_file_md5 == self.query_md5
                        ).filter(Checkpoint.table_name.isnot(None))
            ).group_by(*cols).subquery()

            query = session.query(Checkpoint).join(
                inner_query,
                and_(
                    Checkpoint.project == inner_query.c.project,
                    Checkpoint.commcare == inner_query.c.commcare,
                    Checkpoint.query_file_md5 == inner_query.c.query_file_md5,
                    Checkpoint.table_name == inner_query.c.table_name,
                    Checkpoint.time_of_run == inner_query.c.max_time_of_run
                )
            ).order_by(Checkpoint.table_name.asc())

            # Can't use this since MySQL < 8.0 doesn't support window functions
            # Keeping for future reference
            #
            # window_func = func.row_number().over(
            #     partition_by=Checkpoint.table_name,
            #     order_by=Checkpoint.time_of_run.desc()
            # ).label("row_number")
            # inner_query = self._filter_query(
            #     session.query(Checkpoint, window_func)
            # )
            # inner_query = inner_query.filter(
            #     Checkpoint.query_file_md5 == self.query_md5
            # )
            # inner_query = inner_query.filter(
            #     Checkpoint.table_name.isnot(None)
            # ).subquery()
            #
            # query = session.query(Checkpoint).select_entity_from(inner_query)\
            #     .filter(inner_query.c.row_number == 1)\
            #     .order_by(Checkpoint.table_name.asc())
            #
            return list(query)

    def update_checkpoint(self, run):
        with session_scope(self.Session) as session:
            session.merge(run)

    def _validate_tables(self):
        if not self.table_names:
            raise Exception("Not tables set in checkpoint manager")


class CheckpointManagerWithDetails(object):

    def __init__(self, manager, since_param, pagination_mode):
        self.manager = manager
        self.since_param = since_param
        self.pagination_mode = pagination_mode

    def set_checkpoint(self, checkpoint_time, is_final=False, doc_id=None, cursor=None):
        if self.manager:
            self.manager.set_checkpoint(
                checkpoint_time, self.pagination_mode, is_final, doc_id=doc_id, cursor=cursor
            )


class CheckpointManagerProvider(object):

    def __init__(
        self,
        base_checkpoint_manager=None,
        since=None,
        start_over=None,
    ):
        self.start_over = start_over
        self.since = since
        self.base_checkpoint_manager = base_checkpoint_manager

    def get_since(self, checkpoint_manager):
        if self.start_over:
            return None

        if self.since:
            return self.since

        if checkpoint_manager:
            if checkpoint_manager.data_source == 'ucr':
                last_checkpoint = checkpoint_manager.get_last_checkpoint()
                return last_checkpoint.cursor if last_checkpoint else None

            since = checkpoint_manager.get_time_of_last_checkpoint()
            return dateutil.parser.parse(since) if since else None

    def get_pagination_mode(self, data_source, checkpoint_manager=None):
        """
        Always use the default pagination mode unless we are continuing
        from a previous checkpoint in which case use the same pagination
        mode as before.
        """
        if self.start_over or self.since or not checkpoint_manager:
            return self.get_paginator_for_datasource(data_source)

        last_checkpoint = checkpoint_manager.get_last_checkpoint()
        if not last_checkpoint:
            return self.get_paginator_for_datasource(data_source)

        return last_checkpoint.get_pagination_mode()

    @staticmethod
    def get_paginator_for_datasource(datasource):
        if datasource == 'ucr':
            return PaginationMode.cursor
        return PaginationMode.date_indexed

    def get_checkpoint_manager(self, data_source, table_names):
        """
        This get's called before each table is exported and set in the
        `env`. It is then passed to the API client and used to set the
        checkpoints.

        :param data_source: Data source for this checkout e.g. 'form'
        :param table_names: List of table names being exported to. This
            is a list since multiple tables can be processed by a
            'since' API query.
        """
        manager = None
        if self.base_checkpoint_manager:
            manager = self.base_checkpoint_manager.for_dataset(
                data_source, table_names
            )

        since = self.get_since(manager)
        pagination_mode = self.get_pagination_mode(data_source, checkpoint_manager=manager)
        logger.info(
            "Creating checkpoint manager for tables: %s, since: %s, "
            "pagination_mode: %s",
            #
            ', '.join(table_names),
            since,
            pagination_mode.name,
        )
        if pagination_mode not in PaginationMode.supported_modes():
            logger.warning(
                "\n====================================\n"
                "This export is using a deprecated pagination mode which will "
                "be removed in\n"
                "future versions. To switch to the new mode you must re-sync "
                "your data using\n"
                "`--start-over`. For more details see: %s"
                "\n====================================\n",  #
                "https://github.com/dimagi/commcare-export/releases/tag/1.5.0"
            )
        return CheckpointManagerWithDetails(manager, since, pagination_mode)



import unittest
from datetime import datetime

import requests
import simplejson

import pytest
from commcare_export.checkpoint import CheckpointManagerWithDetails
from commcare_export.commcare_hq_client import (
    CommCareHqClient,
    ResourceRepeatException,
)
from commcare_export.commcare_minilinq import (
    DATE_PARAMS,
    DatePaginator,
    PaginationMode,
    SimplePaginator,
    get_paginator,
)
from mock import Mock, patch

class FakeSession(object):

    def get(self, resource_url, params=None, auth=None, timeout=None):
        result = self._get_results(params)
        # Mutatey construction method required by requests.Response
        response = requests.Response()
        response._content = simplejson.dumps(result).encode('utf-8')
        response.status_code = 200
        return response

    def _get_results(self, params):
        if params:
            assert params['offset'][0] == '1'
            return {
                'meta': {
                    'next': None,
                    'offset': params['offset'][0],
                    'limit': 1,
                    'total_count': 2
                },
                'objects': [{
                    'id': 1,
                    'foo': 2
                }]
            }
        else:
            return {
                'meta': {
                    'next': '?offset=1',
                    'offset': 0,
                    'limit': 1,
                    'total_count': 2
                },
                'objects': [{
                    'id': 2,
                    'foo': 1
                }]
            }


class FakeDateCaseSession(FakeSession):

    def _get_results(self, params):
        if not params:
            return {
                'meta': {
                    'next': '?offset=1',
                    'offset': 0,
                    'limit': 1,
                    'total_count': 2
                },
                'objects': [{
                    'id': 1,
                    'foo': 1,
                    'indexed_on': '2017-01-01T15:36:22Z'
                }]
            }
        else:
            since_query_param = DATE_PARAMS['indexed_on'].start_param
            assert params[since_query_param] == '2017-01-01T15:36:22'
            # include ID=1 again to make sure it gets filtered out
            return {
                'meta': {
                    'next': None,
                    'offset': 1,
                    'limit': 1,
                    'total_count': 2
                },
                'objects': [{
                    'id': 1,
                    'foo': 1
                }, {
                    'id': 2,
                    'foo': 2
                }]
            }


class FakeRepeatedDateCaseSession(FakeSession):
    # Model the case where there are as many or more cases with the same
    # indexed_on than the batch size (2), so the client requests the
    # same set of cases in a loop.
    def _get_results(self, params):
        if not params:
            return {
                'meta': {
                    'next': '?offset=1',
                    'offset': 0,
                    'limit': 2,
                    'total_count': 4
                },
                'objects': [{
                    'id': 1,
                    'foo': 1,
                    'indexed_on': '2017-01-01T15:36:22Z'
                }, {
                    'id': 2,
                    'foo': 2,
                    'indexed_on': '2017-01-01T15:36:22Z'
                }]
            }
        else:
            since_query_param = DATE_PARAMS['indexed_on'].start_param
            assert params[since_query_param] == '2017-01-01T15:36:22'
            return {
                'meta': {
                    'next': '?offset=1',
                    'offset': 0,
                    'limit': 2,
                    'total_count': 4
                },
                'objects': [{
                    'id': 1,
                    'foo': 1,
                    'indexed_on': '2017-01-01T15:36:22Z'
                }, {
                    'id': 2,
                    'foo': 2,
                    'indexed_on': '2017-01-01T15:36:22Z'
                }]
            }


class FakeMessageLogSession(FakeSession):

    def _get_results(self, params):
        obj_1 = {
            'id': 1,
            'foo': 1,
            'date_last_activity': '2017-01-01T15:36:22Z'
        }
        obj_2 = {
            'id': 2,
            'foo': 2,
            'date_last_activity': '2017-01-01T15:37:22Z'
        }
        obj_3 = {
            'id': 3,
            'foo': 3,
            'date_last_activity': '2017-01-01T15:38:22Z'
        }
        if not params:
            return {
                'meta': {
                    'next': '?cursor=xyz',
                    'limit': 2
                },
                'objects': [obj_1, obj_2]
            }
        else:
            since_query_param = DATE_PARAMS['date_last_activity'].start_param
            since = params[since_query_param]
            if since == '2017-01-01T15:37:22':
                return {
                    'meta': {
                        'next': '?cursor=xyz',
                        'limit': 2
                    },
                    'objects': [obj_3]
                }
            if since == '2017-01-01T15:38:22':
                return {'meta': {'next': None, 'limit': 2}, 'objects': []}

            raise Exception(since)


class FakeDateFormSession(FakeSession):

    def _get_results(self, params):
        since1 = '2017-01-01T15:36:22'
        since2 = '2017-01-01T16:00:00'
        if not params:
            return {
                'meta': {
                    'next': '?offset=1',
                    'offset': 0,
                    'limit': 1,
                    'total_count': 3
                },
                'objects': [{
                    'id': 1,
                    'foo': 1,
                    'indexed_on': '{}Z'.format(since1)
                }]
            }
        else:
            since_query_param = DATE_PARAMS['indexed_on'].start_param
            indexed_on = params[since_query_param]
            if indexed_on == since1:
                # include ID=1 again to make sure it gets filtered out
                return {
                    'meta': {
                        'next': '?offset=2',
                        'offset': 0,
                        'limit': 1,
                        'total_count': 3
                    },
                    'objects': [{
                        'id': 1,
                        'foo': 1
                    }, {
                        'id': 2,
                        'foo': 2,
                        'indexed_on': '{}Z'.format(since2)
                    }]
                }
            elif indexed_on == since2:
                return {
                    'meta': {
                        'next': None,
                        'offset': 0,
                        'limit': 1,
                        'total_count': 3
                    },
                    'objects': [{
                        'id': 3,
                        'foo': 3
                    }]
                }
            else:
                raise Exception(indexed_on)


class TestCommCareHqClient(unittest.TestCase):

    def _test_iterate(self, session, paginator, expected_count, expected_vals):
        client = CommCareHqClient(
            '/fake/commcare-hq/url', 'fake-project', None, None
        )
        client.session = session

        # Iteration should do two "gets" because the first will have
        # something in the "next" metadata field
        paginator.init()
        checkpoint_manager = CheckpointManagerWithDetails(
            None, None, PaginationMode.date_indexed
        )
        results = list(
            client.iterate(
                '/fake/uri', paginator, checkpoint_manager=checkpoint_manager
            )
        )
        self.assertEqual(len(results), expected_count)
        self.assertEqual([result['foo'] for result in results], expected_vals)

    def test_iterate_simple(self):
        self._test_iterate(FakeSession(), SimplePaginator('fake'), 2, [1, 2])

    def test_iterate_date(self):
        self._test_iterate(
            FakeDateFormSession(), get_paginator('form'), 3, [1, 2, 3]
        )
        self._test_iterate(
            FakeDateCaseSession(), get_paginator('case'), 2, [1, 2]
        )

    def test_repeat_limit(self):
        with pytest.raises(
            ResourceRepeatException,
            match="Requested resource '/fake/uri' 10 times with same parameters"
        ):
            self._test_iterate(
                FakeRepeatedDateCaseSession(), get_paginator('case', 2), 2,
                [1, 2]
            )

    def test_message_log(self):
        self._test_iterate(
            FakeMessageLogSession(), get_paginator('messaging-event', 2), 3,
            [1, 2, 3]
        )

    @patch("commcare_export.commcare_hq_client.CommCareHqClient.session")
    def test_dont_raise_on_too_many_requests(self, session_mock):
        response = requests.Response()
        response.headers = {'Retry-After': "0.0"}
        client = CommCareHqClient(
            '/fake/commcare-hq/url', 'fake-project', None, None
        )

        self.assertFalse(client._should_raise_for_status(response))

    @patch("commcare_export.commcare_hq_client.CommCareHqClient.session")
    def test_raise_on_too_many_requests(self, session_mock):
        response = requests.Response()
        response.headers = {}

        client = CommCareHqClient(
            '/fake/commcare-hq/url', 'fake-project', None, None
        )

        self.assertTrue(client._should_raise_for_status(response))

class TestDatePaginator(unittest.TestCase):

    @classmethod
    def setup_class(cls):
        pass

    def test_empty_batch(self):
        self.assertIsNone(
            DatePaginator('since', params=SimplePaginator()
                         ).next_page_params_from_batch({'objects': []})
        )

    def test_bad_date(self):
        self.assertIsNone(
            DatePaginator('since', params=SimplePaginator()
                         ).next_page_params_from_batch({
                             'objects': [{
                                 'since': 'not a date'
                             }]
                         })
        )

    def test_multi_field_sort(self):
        d1 = '2017-01-01T15:36:22Z'
        d2 = '2017-01-01T18:36:22Z'
        paginator = DatePaginator(['s1', 's2'], params=SimplePaginator())
        self.assertEqual(
            paginator.get_since_date({'objects': [{
                's1': d1,
                's2': d2
            }]}), datetime.strptime(d1, '%Y-%m-%dT%H:%M:%SZ')
        )

        self.assertEqual(
            paginator.get_since_date({'objects': [{
                's2': d2
            }]}), datetime.strptime(d2, '%Y-%m-%dT%H:%M:%SZ')
        )

        self.assertEqual(
            paginator.get_since_date({'objects': [{
                's1': None,
                's2': d2
            }]}), datetime.strptime(d2, '%Y-%m-%dT%H:%M:%SZ')
        )



import datetime
import uuid

import sqlalchemy

import pytest
from commcare_export.checkpoint import (
    Checkpoint,
    CheckpointManager,
    CheckpointManagerProvider,
    session_scope,
)
from commcare_export.commcare_minilinq import PaginationMode


@pytest.fixture()
def manager(db_params):
    manager = CheckpointManager(
        db_params['url'],
        'query',
        '123',
        'test',
        'hq',
        poolclass=sqlalchemy.pool.NullPool
    )
    try:
        yield manager
    finally:
        with manager:
            manager.connection.execute(
                sqlalchemy.sql
                .text('DROP TABLE IF EXISTS commcare_export_runs')
            )
            manager.connection.execute(
                sqlalchemy.sql.text('DROP TABLE IF EXISTS alembic_version')
            )


@pytest.fixture()
def configured_manager(manager):
    manager.create_checkpoint_table()
    return manager


@pytest.mark.dbtest
class TestCheckpointManager(object):

    def test_create_checkpoint_table(self, manager, revision='head'):
        manager.create_checkpoint_table(revision)
        with manager:
            table = manager.get_table('commcare_export_runs')
            assert table is not None

    def test_checkpoint_table_exists(self, manager):
        # Test that the migrations don't fail for tables that existed before
        # migrations were used.
        # This test can be removed at some point in the future.
        self.test_create_checkpoint_table(manager, '9945abb4ec70')
        with manager:
            manager.connection.execute(
                sqlalchemy.sql.text('DROP TABLE alembic_version')
            )
        manager.create_checkpoint_table()

    def test_get_time_of_last_checkpoint(self, configured_manager):
        manager = configured_manager.for_dataset('form', ['t1'])
        manager.set_checkpoint(
            datetime.datetime.utcnow(), PaginationMode.date_indexed
        )
        second_run = datetime.datetime.utcnow()
        manager.set_checkpoint(second_run, PaginationMode.date_indexed)

        assert manager.get_time_of_last_checkpoint() == second_run.isoformat()

    def test_get_last_checkpoint_no_args(self, configured_manager):
        # test that we can still get the time of last run no project and commcare args
        with session_scope(configured_manager.Session) as session:
            since_param = datetime.datetime.utcnow().isoformat()
            session.add(
                Checkpoint(
                    id=uuid.uuid4().hex,
                    query_file_name=configured_manager.query,
                    query_file_md5=configured_manager.query_md5,
                    project=None,
                    commcare=None,
                    since_param=since_param,
                    time_of_run=datetime.datetime.utcnow().isoformat(),
                    final=True
                )
            )
        manager = configured_manager.for_dataset('form', ['t1', 't2'])
        checkpoint = manager.get_last_checkpoint()
        assert checkpoint.since_param == since_param
        assert checkpoint.project == manager.project
        assert checkpoint.commcare == manager.commcare
        assert len(manager.get_latest_checkpoints()) == 2

    def test_get_last_checkpoint_no_table(self, configured_manager):
        # test that we can still get the time of last run no table
        # also tests that new checkoints are created with the tables
        with session_scope(configured_manager.Session) as session:
            since_param = datetime.datetime.utcnow().isoformat()
            session.add(
                Checkpoint(
                    id=uuid.uuid4().hex,
                    query_file_name=configured_manager.query,
                    query_file_md5=configured_manager.query_md5,
                    project=None,
                    commcare=None,
                    since_param=since_param,
                    time_of_run=datetime.datetime.utcnow().isoformat(),
                    final=True
                )
            )

            session.add(
                Checkpoint(
                    id=uuid.uuid4().hex,
                    query_file_name=configured_manager.query,
                    query_file_md5=configured_manager.query_md5,
                    project=configured_manager.project,
                    commcare=configured_manager.commcare,
                    since_param=since_param,
                    time_of_run=datetime.datetime.utcnow().isoformat(),
                    final=True
                )
            )
        manager = configured_manager.for_dataset('form', ['t1', 't2'])
        checkpoint = manager.get_last_checkpoint()
        assert checkpoint.since_param == since_param
        assert checkpoint.table_name in manager.table_names
        checkpoints = manager.get_latest_checkpoints()
        assert len(checkpoints) == 2
        assert {c.table_name for c in checkpoints} == set(manager.table_names)

    def test_clean_on_final_run(self, configured_manager):
        manager = configured_manager.for_dataset('form', ['t1'])
        manager.set_checkpoint(
            datetime.datetime.utcnow(),
            PaginationMode.date_indexed,
            doc_id="1"
        )
        manager.set_checkpoint(
            datetime.datetime.utcnow(),
            PaginationMode.date_indexed,
            doc_id="2"
        )

        def _get_non_final_rows_count():
            with session_scope(manager.Session) as session:
                return session.query(Checkpoint).filter_by(final=False).count()

        assert _get_non_final_rows_count() == 2
        manager.set_checkpoint(
            datetime.datetime.utcnow(),
            PaginationMode.date_indexed,
            True,
            doc_id="3"
        )
        assert _get_non_final_rows_count() == 0

    def test_get_time_of_last_checkpoint_with_key(self, configured_manager):
        manager = configured_manager.for_dataset('form', ['t1'])
        manager.key = 'my key'
        last_run_time = datetime.datetime.utcnow()
        manager.set_checkpoint(last_run_time, PaginationMode.date_indexed)

        assert manager.get_time_of_last_checkpoint(
        ) == last_run_time.isoformat()
        manager.key = None
        assert manager.get_time_of_last_checkpoint() is None

    def test_multiple_tables(self, configured_manager):
        t1 = uuid.uuid4().hex
        t2 = uuid.uuid4().hex
        manager = configured_manager.for_dataset('form', [t1, t2])
        last_run_time = datetime.datetime.utcnow()
        doc_id = uuid.uuid4().hex
        manager.set_checkpoint(
            last_run_time, PaginationMode.date_indexed, doc_id=doc_id
        )

        assert manager.for_dataset('form', [
            t1
        ]).get_time_of_last_checkpoint() == last_run_time.isoformat()
        assert manager.for_dataset('form', [
            t2
        ]).get_time_of_last_checkpoint() == last_run_time.isoformat()
        assert manager.for_dataset('form',
                                   ['t3']).get_last_checkpoint() is None

        checkpoints = manager.list_checkpoints()
        assert len(checkpoints) == 2
        assert {checkpoints[0].table_name,
                checkpoints[1].table_name} == {t1, t2}
        assert {checkpoints[0].last_doc_id,
                checkpoints[1].last_doc_id} == {doc_id}

    def test_get_latest_checkpoints(self, configured_manager):
        manager = configured_manager.for_dataset('form', ['t1', 't2'])
        manager.set_checkpoint(
            datetime.datetime.utcnow(), PaginationMode.date_indexed
        )

        manager.query_md5 = '456'
        manager.set_checkpoint(
            datetime.datetime.utcnow(), PaginationMode.date_indexed
        )
        latest_time = datetime.datetime.utcnow()
        manager.set_checkpoint(latest_time, PaginationMode.date_indexed)

        checkpoints = manager.get_latest_checkpoints()
        assert len(checkpoints) == 2
        assert [c.table_name for c in checkpoints] == ['t1', 't2']
        assert {c.query_file_md5 for c in checkpoints} == {'456'}
        assert {c.since_param for c in checkpoints
               } == {latest_time.isoformat()}


@pytest.mark.parametrize(
    'since, start_over, expected_since, expected_paginator', [
        (None, True, None, PaginationMode.date_indexed),
        ('since', False, 'since', PaginationMode.date_indexed),
        (None, False, None, PaginationMode.date_indexed),
    ]
)
def test_checkpoint_details_static(
    since,
    start_over,
    expected_since,
    expected_paginator,
):
    cmp = CheckpointManagerProvider(None, since, start_over)
    assert expected_since == cmp.get_since(None)
    assert expected_paginator == cmp.get_pagination_mode('', None)


@pytest.mark.dbtest
class TestCheckpointManagerProvider(object):

    def test_checkpoint_details_no_checkpoint(self, configured_manager):
        manager = configured_manager.for_dataset('form', ['t1'])
        assert None is CheckpointManagerProvider().get_since(manager)
        assert PaginationMode.date_indexed == CheckpointManagerProvider(
        ).get_pagination_mode('form', manager)

    def test_checkpoint_details_latest_from_db(self, configured_manager):
        data_source = 'form'
        manager = configured_manager.for_dataset(data_source, ['t1'])

        self._test_checkpoint_details(
            manager, datetime.datetime.utcnow(), PaginationMode.date_modified, data_source
        )
        self._test_checkpoint_details(
            manager, datetime.datetime.utcnow(), PaginationMode.date_indexed, data_source
        )
        self._test_checkpoint_details(
            manager, datetime.datetime.utcnow(), PaginationMode.date_modified, data_source
        )

    def _test_checkpoint_details(
        self,
        manager,
        checkpoint_date,
        pagination_mode,
        data_source,
    ):
        manager.set_checkpoint(checkpoint_date, pagination_mode)

        cmp = CheckpointManagerProvider()
        assert pagination_mode == cmp.get_pagination_mode(data_source, manager)
        assert checkpoint_date == cmp.get_since(manager)



"""Added cursor to checkpoint

Revision ID: 3b37b3b06104
Revises: 6f158d161ab6
Create Date: 2023-08-25 11:10:38.713189

"""
from alembic import op
import sqlalchemy as sa


revision = '3b37b3b06104'
down_revision = '6f158d161ab6'
branch_labels = None
depends_on = None


def upgrade():
    url = op.get_bind().engine.url
    collation = 'utf8_bin' if 'mysql' in url.drivername else None
    op.add_column(
        'commcare_export_runs',
        sa.Column('cursor', sa.Unicode(255, collation=collation))
    )

def downgrade():
    op.drop_column('commcare_export_runs', 'cursor')



from jsonpath_ng import jsonpath


def split_leftmost(jsonpath_expr):
    if isinstance(jsonpath_expr, jsonpath.Child):
        further_leftmost, rest = split_leftmost(jsonpath_expr.left)
        return further_leftmost, rest.child(jsonpath_expr.right)
    elif isinstance(jsonpath_expr, jsonpath.Descendants):
        further_leftmost, rest = split_leftmost(jsonpath_expr.left)
        return further_leftmost, jsonpath.Descendants(
            rest, jsonpath_expr.right
        )
    else:
        return jsonpath_expr, jsonpath.This()



from __future__ import (
    absolute_import,
    division,
    generators,
    nested_scopes,
    print_function,
    unicode_literals,
)

from collections import defaultdict, namedtuple

from commcare_export.exceptions import (
    LongFieldsException,
    MissingColumnException,
    ReservedTableNameException,
)
from commcare_export.jsonpath_utils import split_leftmost
from commcare_export.map_format import compile_map_format_via
from commcare_export.minilinq import *
from jsonpath_ng import jsonpath
from jsonpath_ng.parser import parse as parse_jsonpath


def take_while(pred, iterator):
    for v in iterator:
        if pred(v):
            yield v
        else:
            return


def drop_while(pred, iterator):
    for v in iterator:
        if not pred(v):
            yield v
            break

    for v in iterator:
        yield v


def without_empty_tail(cells):
    """
    Returns the prefix of a column that is not entirely empty.
    """
    return list(
        reversed(
            list(
                drop_while(
                    lambda v: (not v) or (not v.value), reversed(cells)
                )
            )
        )
    )


def map_value(mappings_sheet, mapping_name, source_value):
    """
    From the mappings_sheet, replaces the source_value with appropriate
    output value
    """
    return source_value


def get_column_by_name(worksheet, column_name):
    # columns and rows are indexed from 1
    for col in range(1, worksheet.max_column + 1):
        value = worksheet.cell(row=1, column=col).value
        value = value.lower().strip() if value else value
        if column_name == value:
            return without_empty_tail([
                worksheet.cell(row=i, column=col)
                for i in range(2, worksheet.max_row + 1)
            ])


def get_columns_by_prefix(worksheet, column_prefix):
    # columns and rows are indexed from 1
    for col in range(1, worksheet.max_column + 1):
        value = worksheet.cell(row=1, column=col).value
        if value and value.lower().startswith(column_prefix):
            yield value, without_empty_tail([
                worksheet.cell(row=i, column=col)
                for i in range(2, worksheet.max_row + 1)
            ])


def compile_mappings(worksheet):
    mapping_names = get_column_by_name(worksheet, "mapping name")
    sources = extended_to_len(
        len(mapping_names), get_column_by_name(worksheet, "source")
    )
    destinations = extended_to_len(
        len(mapping_names), get_column_by_name(worksheet, "destination")
    )

    mappings = defaultdict(lambda: defaultdict(lambda: None))

    for mapping_name, source, dest in zip(
        mapping_names, sources, destinations
    ):
        if mapping_name and source:
            mappings[mapping_name.value][source.value
                                        ] = dest.value if dest else None

    return mappings


def compile_filters(worksheet, mappings=None):
    filter_names = [
        cell.value
        for cell in get_column_by_name(worksheet, 'filter name') or []
    ]

    if not filter_names:
        return []

    filter_values = extended_to_len(
        len(filter_names),
        [
            cell.value
            for cell in get_column_by_name(worksheet, 'filter value') or []
        ]
    )
    # Preserve values of duplicate filter names. Results in an OR filter.
    # e.g. {'type': ['person'], 'owner_id': ['abc123', 'def456']}
    filters = defaultdict(list)
    for k, v in zip(filter_names, filter_values):
        filters[k].append(v)
    return filters


def extended_to_len(desired_len, some_list, value=None):
    return [
        some_list[i] if i < len(some_list) else value
        for i in range(0, desired_len)
    ]


def _get_safe_source_field(source_field):

    def _safe_node(node):
        try:
            parse_jsonpath(node)
        except Exception:
            # quote nodes with special characters
            return '"{}"'.format(node)
        else:
            return node

    try:
        parse_jsonpath(source_field)
    except Exception:
        source_field = '.'.join([
            _safe_node(node) if node else node
            for node in source_field.split('.')
        ])
        if source_field.endswith('.'):
            raise Exception("Blank node path: {}".format(source_field))

    return Reference(source_field)


def compile_field(
    field,
    source_field,
    alternate_source_fields=None,
    map_via=None,
    format_via=None,
    mappings=None
):
    expr = _get_safe_source_field(source_field)

    if alternate_source_fields:
        expr = Apply(
            Reference('or'),
            expr,
            *[Reference(alt_field) for alt_field in alternate_source_fields]
        )
    if map_via:
        expr = compile_map_format_via(expr, map_via)

    if format_via:
        expr = compile_map_format_via(expr, format_via)

    if mappings and field in mappings:
        expr = compile_mapped_field(mappings[field], expr)

    return expr


def compile_mapped_field(field_mappings, field_expression):
    # quote the ref in case it has special chars
    quoted_field = Apply(
        Reference('join'),
        Literal(''),
        Literal('"'),
        field_expression,
        Literal('"')
    )
    # produce the mapping reference i.e. 'mapping."X"'
    mapping_ref = Apply(
        Reference('join'), Literal('.'), Literal('mapping'), quoted_field
    )
    # apply the reference to the field mappings to get the final value
    mapped_value = FlatMap(
        source=Literal([field_mappings]),
        body=Reference(mapping_ref),
        name='mapping'
    )
    return Apply(Reference('default'), mapped_value, field_expression)


def _get_alternate_source_fields_from_csv(worksheet, num_fields):

    def _clean_csv_field(field):
        if field and field.value:
            return [val.strip() for val in field.value.split(',')]

    alt_source_col = get_column_by_name(worksheet, 'alternate source fields')

    if alt_source_col:
        alt_source_fields = extended_to_len(num_fields, alt_source_col)
        return [_clean_csv_field(field) for field in alt_source_fields]


def _get_alternate_source_fields_from_columns(worksheet, num_fields):
    matching_columns = sorted(
        get_columns_by_prefix(worksheet, 'alternate source field'),
        key=lambda x: x[0]
    )
    alt_source_cols = [
        extended_to_len(
            num_fields, [cell.value if cell else cell for cell in alt_col]
        ) for (col_name, alt_col) in matching_columns
    ]
    # transpose columns to rows
    alt_srouce_fields = map(list, zip(*alt_source_cols))
    return [list(filter(None, fields)) for fields in alt_srouce_fields]


def get_alternate_source_fields(worksheet, num_fields):
    return (
        _get_alternate_source_fields_from_csv(worksheet, num_fields)
        or _get_alternate_source_fields_from_columns(worksheet, num_fields)
        or extended_to_len(num_fields, [])
    )


def compile_fields(worksheet, mappings=None):
    fields = without_empty_tail(get_column_by_name(worksheet, 'field') or [])

    if not fields:
        return []

    source_fields = extended_to_len(
        len(fields),
        get_column_by_name(worksheet, 'source field') or []
    )
    map_vias = extended_to_len(
        len(fields),
        get_column_by_name(worksheet, 'map via') or []
    )
    format_vias = extended_to_len(
        len(fields),
        get_column_by_name(worksheet, 'format via') or []
    )

    alternate_source_fields = get_alternate_source_fields(
        worksheet, len(fields)
    )

    args = zip(
        fields, source_fields, alternate_source_fields, map_vias, format_vias
    )
    return [
        compile_field(
            field=field.value,
            source_field=source_field.value,
            alternate_source_fields=alt_source_fields,
            map_via=map_via.value if map_via else None,
            format_via=format_via.value if format_via else None,
            mappings=mappings
        )
        for (field, source_field, alt_source_fields, map_via, format_via)
        in args
    ]


def compile_source(worksheet, value_or_root=False):
    """
    Compiles just the part of the Excel Spreadsheet that indicates the
    API endpoint to hit along with optional filters and an optional
    JSONPath within that endpoint,

    For example, this spreadsheet

    Data Source                    Filter Name   Filter Value        Include Referenced Items
    -----------------------------  ------------  ------------------  --------------------------
    form[*].form.child_questions   app_id        <app id>            cases
                                   xmlns.exact   <some form xmlns>

    Should fetch from api/form?app_id=<app id>&xmlns.exact=<some form
    xmlns>&cases__full=true and then iterate (FlatMap) over all child
    questions.

    :return: tuple of the 'data source' expression and the 'root doc
        expression'.

        'data source': The MiniLinq that calls 'api_data' function to
        get data from CommCare

        'root doc expression': The MiniLinq that is applied to each doc,
        can be None.
    """

    data_source_column = get_column_by_name(worksheet, 'data source')
    if not data_source_column:
        raise Exception('Sheet has no "Data Source" column.')
    data_source_str = data_source_column[0].value
    filters = compile_filters(worksheet)
    include_referenced_items = [
        cell.value for cell in
        (get_column_by_name(worksheet, 'include referenced items') or [])
    ]

    data_source, data_source_jsonpath = split_leftmost(
        parse_jsonpath(data_source_str)
    )
    maybe_redundant_slice, remaining_jsonpath = split_leftmost(
        data_source_jsonpath
    )

    # The leftmost _must_ be of type Fields with one field and will pull
    # out the first field
    if not isinstance(data_source,
                      jsonpath.Fields) or len(data_source.fields) > 1:
        raise Exception('Bad value for data source: %s' % str(data_source))

    data_source = data_source.fields[0]

    if isinstance(maybe_redundant_slice, jsonpath.Slice):
        data_source_jsonpath = remaining_jsonpath

    api_query_args = [
        Reference("api_data"),
        Literal(data_source),
        Reference('checkpoint_manager')
    ]

    if not filters:
        if include_referenced_items:
            # Pad the argument list if we have further args; keeps tests
            # and user code more readable at the expense of this
            # conditional
            api_query_args.append(Literal(None))
    else:
        api_query_args.append(Literal(filters))

    if include_referenced_items:
        api_query_args.append(Literal(include_referenced_items))

    api_query = Apply(*api_query_args)

    if (
        data_source_jsonpath is None
        or isinstance(data_source_jsonpath, jsonpath.This)
        or isinstance(data_source_jsonpath, jsonpath.Root)
    ):
        return data_source, api_query, None
    else:
        if value_or_root:
            # if the jsonpath doesn't yield a value yield the root document
            expr = get_value_or_root_expression(data_source_jsonpath)
        else:
            expr = Reference(str(data_source_jsonpath))
        return data_source, api_query, expr


def get_value_or_root_expression(value_expression):
    """
    Return expression used when iterating over a nested document but
    also wanting a record if the value expression returns an empty
    result.
    """

    # We add a bind here so that in JsonPathEnv we can restrict
    # expressions to only those that reference the root. That prevents
    # us from mistakenly getting values from the root that happen to
    # have the same name as those in the child.
    root_expr = Bind("__root_only", Literal(True), Reference("$"))
    return Apply(
        Reference('_or_raw'), Reference(str(value_expression)), root_expr
    )


# If the source is expected to provide a column, then require that it is
# already present or can be added without conflicting with an existing
# column.
def require_column_in_sheet(
    sheet_name,
    data_source,
    table_name,
    output_headings,
    output_fields,
    column_enforcer,
):
    # Check for conflicting use of column name.
    extend_fields = True

    required_column = column_enforcer.column_to_require(data_source)
    if required_column is None:
        extend_fields = False
    else:
        for i in range(len(output_headings)):
            if output_headings[i].value == required_column.name.v:
                if isinstance(output_fields[i], Reference) and \
                        output_fields[i].ref == required_column.source:
                    extend_fields = False
                    continue
                else:
                    raise Exception(
                        'Field name "{}" conflicts with an internal name.'
                        .format(required_column.name.v)
                    )

    if extend_fields:
        headings = [
            Literal(output_heading.value) for output_heading in output_headings
        ] + [required_column.name]
        body = List(
            output_fields + [
                compile_field(
                    field=required_column.name,
                    source_field=required_column.source
                )
            ]
        )
    else:
        headings = [
            Literal(output_heading.value) for output_heading in output_headings
        ]
        body = List(output_fields)

    return (headings, body)


def parse_sheet(
    worksheet,
    mappings=None,
    column_enforcer=None,
    value_or_root=False,
):
    mappings = mappings or {}
    data_source, source_expr, root_doc_expr = compile_source(
        worksheet, value_or_root
    )

    table_name_column = get_column_by_name(worksheet, 'table name')
    if table_name_column:
        output_table_name = table_name_column[0].value
    else:
        output_table_name = worksheet.title
    output_headings = get_column_by_name(worksheet, 'field')
    output_types = get_column_by_name(worksheet, 'data type') or []
    output_fields = compile_fields(worksheet, mappings=mappings)

    if not output_fields:
        headings = []
        data_types = []
        source = source_expr
        body = None
    else:
        # note: if we want to add data types to the columns added by the
        # column_enforcer this will have to conditionally move into the
        # if/else below
        data_types = [Literal(data_type.value) for data_type in output_types]
        if column_enforcer is not None:
            (headings, body) = require_column_in_sheet(
                worksheet.title,
                data_source,
                output_table_name,
                output_headings,
                output_fields,
                column_enforcer
            )
            source = source_expr
        else:
            headings = [
                Literal(output_heading.value)
                for output_heading in output_headings
            ]
            source = source_expr
            body = List(output_fields)

    return SheetParts(
        output_table_name,
        headings,
        source,
        body,
        root_doc_expr,
        data_types,
        data_source,
    )


class SheetParts(namedtuple(
    'SheetParts',
    'name headings source body root_expr data_types data_source'
)):

    def __new__(
        cls,
        name,
        headings,
        source,
        body,
        root_expr=None,
        data_types=None,
        data_source=None
    ):
        return super().__new__(
            cls,
            name,
            headings,
            source,
            body,
            root_expr,
            data_types or [],
            data_source
        )

    @property
    def columns(self):
        return [col.v for col in self.headings]


def parse_workbook(workbook, column_enforcer=None, value_or_root=False):
    """
    Returns a MiniLinq corresponding to the Excel configuration, which
    consists of the following sheets:

    1. "Mappings" a sheet with three columns that defines simple lookup
       table functions:
       A. MappingName - the name by which this mapping is referenced
       B. Source - the value to match
       C. Destination - the value to return

    2. Each other sheet represents one data table to emit
    """
    try:
        mappings_sheet = workbook['Mappings']
    except KeyError:
        mappings_sheet = None
    mappings = compile_mappings(mappings_sheet) if mappings_sheet else None

    emit_sheets = [
        sheet_name for sheet_name in workbook.sheetnames
        if sheet_name != 'Mappings'
    ]

    parsed_sheets = []
    for sheet in emit_sheets:
        try:
            sheet_parts = parse_sheet(
                workbook[sheet], mappings, column_enforcer, value_or_root
            )
        except Exception as e:
            msg = 'Ignoring sheet "{}": {}'.format(sheet, str(e))
            if logger.isEnabledFor(logging.DEBUG):
                logger.exception(msg)
            else:
                logger.warning(msg)
            continue

        parsed_sheets.append(sheet_parts)

    return parsed_sheets


def compile_queries(parsed_sheets, missing_value, combine_emits):
    # group sheets by source
    sheets_by_source = []
    for sheet in parsed_sheets:
        # Not easy to implement hashing on MiniLinq objects so can't use
        # a dict
        for source, sheets in sheets_by_source:
            if sheet.source == source:
                sheets.append(sheet)
                break
        else:
            sheets_by_source.append((sheet.source, [sheet]))

    queries = []
    for source, sheets in sheets_by_source:
        if len(sheets) > 1:
            if combine_emits:
                queries.append(
                    get_multi_emit_query(source, sheets, missing_value)
                )
            else:
                queries.extend([
                    get_single_emit_query(sheet, missing_value)
                    for sheet in sheets
                ])
        else:
            queries.append(get_single_emit_query(sheets[0], missing_value))
    return queries


def get_multi_emit_query(source, sheets, missing_value):
    """
    Multiple `Emit` expressions using the same data source. For this we
    reverse the `Map` so that we apply each `Emit` repeatedly for each
    doc produced by the data source.
    """
    emits = []
    # the filter here is to prevent accumulating a `[None]` value for
    # each doc
    multi_query = Filter(
        predicate=Apply(Reference("filter_empty"), Reference("$")),
        source=Map(source=source, body=List(emits))
    )

    for sheet in sheets:
        # if there is no root expression then we just reference the
        # whole document with `this`
        root_expr = sheet.root_expr or Reference("`this`")
        emits.append(
            Emit(
                table=sheet.name,
                headings=sheet.headings,
                source=Map(source=root_expr, body=sheet.body),
                missing_value=missing_value,
                data_types=sheet.data_types,
            )
        )

    table_names = [sheet.name for sheet in sheets]
    data_source = sheets[
        0].data_source  # sheets will all have the same datasource
    return Bind(
        'checkpoint_manager',
        Apply(
            Reference('get_checkpoint_manager'),
            Literal(data_source),
            Literal(table_names)
        ),
        multi_query
    )


def get_single_emit_query(sheet, missing_value):
    """
    Single `Emit` for the data source to we can just apply the `Emit`
    once with the source expression being the data source.
    """

    def _get_source(source, root_expr):
        if root_expr:
            return FlatMap(source=source, body=root_expr)
        else:
            return source

    emit = Emit(
        table=sheet.name,
        headings=sheet.headings,
        source=Map(
            source=_get_source(sheet.source, sheet.root_expr), body=sheet.body
        ),
        missing_value=missing_value,
        data_types=sheet.data_types,
    )
    return Bind(
        'checkpoint_manager',
        Apply(
            Reference('get_checkpoint_manager'),
            Literal(sheet.data_source),
            Literal([sheet.name])
        ),
        emit
    )


def check_field_length(parsed_sheets, max_column_length):
    long_fields = defaultdict(list)
    for sheet in parsed_sheets:
        for col in sheet.columns:
            if len(col) > max_column_length:
                long_fields[sheet.name].append(col)

    if long_fields:
        raise LongFieldsException(long_fields, max_column_length)


def check_columns(parsed_sheets, columns):
    columns = set(columns)
    errors_by_sheet = {}
    for sheet in parsed_sheets:
        missing = columns - set(sheet.columns)
        if missing:
            errors_by_sheet[sheet.name] = list(missing)
    if errors_by_sheet:
        raise MissingColumnException(errors_by_sheet)


blacklisted_tables = []


def blacklist(table_name):
    blacklisted_tables.append(table_name)


def get_queries_from_excel(
    workbook,
    missing_value=None,
    combine_emits=False,
    max_column_length=None,
    required_columns=None,
    column_enforcer=None,
    value_or_root=False
):
    parsed_sheets = parse_workbook(workbook, column_enforcer, value_or_root)
    for sheet in parsed_sheets:
        if sheet.name in blacklisted_tables:
            raise ReservedTableNameException(sheet.name)
    if max_column_length:
        check_field_length(parsed_sheets, max_column_length)
    if required_columns:
        check_columns(parsed_sheets, required_columns)
    queries = compile_queries(parsed_sheets, missing_value, combine_emits)
    return List(queries) if len(queries) > 1 else queries[0]



import logging

from commcare_export.commcare_minilinq import SimplePaginator
from commcare_export.misc import unwrap_val

logger = logging.getLogger(__name__)

# LocationInfoProvider uses the /location_type/ endpoint of the API to
# retrieve location type data, stores that information in a dictionary
# keyed by resource URI and provides the method 'get_location_info' to
# extract values from the dictionary.


class LocationInfoProvider:

    def __init__(self, api_client, page_size):
        self._api_client = api_client
        self._page_size = page_size
        self._location_types = None
        self._location_hierarchy = None

    @property
    def location_types(self):
        if self._location_types is None:
            self._location_types = self.get_location_types()
        return self._location_types

    def get_location_info(self, resource_uri, field):
        unwrapped_uri = unwrap_val(resource_uri)
        if unwrapped_uri in self.location_types:
            location_type = self.location_types[unwrapped_uri]
            if field in self.location_types[unwrapped_uri]:
                return location_type[field]
        return None

    def get_location_ancestor(self, resource_uri, location_type_code):
        unwrapped_uri = unwrap_val(resource_uri)
        if unwrapped_uri in self.location_hierarchy:
            location_hierarchy = self.location_hierarchy[unwrapped_uri]
            if location_type_code in location_hierarchy:
                return location_hierarchy[location_type_code]
        return None

    def get_location_types(self):
        paginator = SimplePaginator('location_type', self._page_size)
        paginator.init(None, False, None)
        location_type_dict = {}
        for row in self._api_client.iterate(
            'location_type', paginator, {'limit': self._page_size}
        ):
            location_type_dict[row['resource_uri']] = row
        return location_type_dict

    @property
    def location_hierarchy(self):
        if self._location_hierarchy is None:
            self._location_hierarchy = self.get_location_hierarchy()
        return self._location_hierarchy

    def get_location_hierarchy(self):
        paginator = SimplePaginator('location', self._page_size)
        paginator.init(None, False, None)

        # Extract every location, its type and its parent
        location_data = {}
        for row in self._api_client.iterate(
            'location', paginator, {'limit': self._page_size}
        ):
            location_data[row['resource_uri']] = {
                'location_id': row['location_id'],
                'location_type': row['location_type'],
                'parent': row['parent'] if 'parent' in row else None
            }

        # Build a map from location resource_uri to a map from
        # location_type_code to ancestor location id.
        ancestors = {}  # includes location itself
        for resource_uri in location_data:
            loc_uri = resource_uri
            type_code_to_id = {}
            while loc_uri is not None:
                if loc_uri not in location_data:
                    logger.warning(
                        'Unknown location referenced: {}'.format(loc_uri)
                    )
                    break

                loc_data = location_data[loc_uri]
                loc_type = loc_data['location_type']
                if loc_type not in self.location_types:
                    logger.warning(
                        'Unknown location type referenced: {}'
                        .format(loc_type)
                    )
                    break

                type_code = self.location_types[loc_type]['code']
                type_code_to_id[type_code] = loc_data['location_id']
                loc_uri = loc_data['parent']
            ancestors[resource_uri] = type_code_to_id
        return ancestors



"""
CommCare/Export-specific extensions to MiniLinq.

To date, this is simply built-ins for querying the
API directly.
"""
import json
import logging
from enum import Enum
from urllib.parse import parse_qs, urlparse
from datetime import datetime

from dateutil.parser import ParserError, parse

from commcare_export.env import CannotBind, CannotReplace, DictEnv
from commcare_export.misc import unwrap

logger = logging.getLogger(__name__)

SUPPORTED_RESOURCES = {
    'form',
    'case',
    'user',
    'location',
    'application',
    'web-user',
    'messaging-event',
    'ucr',
}

DEFAULT_PAGE_SIZE = 1000
DEFAULT_UCR_PAGE_SIZE = 10000


class PaginationMode(Enum):
    date_indexed = "date_indexed"
    date_modified = "date_modified"
    cursor = "cursor"

    @classmethod
    def supported_modes(cls):
        return [
            cls.date_indexed,
            cls.cursor,
        ]


class SimpleSinceParams(object):

    def __init__(self, start, end):
        self.start_param = start
        self.end_param = end

    def __call__(self, since, until):
        params = {}
        if since:
            params[self.start_param] = since.isoformat()
        if until:
            params[self.end_param] = until.isoformat()
        return params


class FormFilterSinceParams(object):

    def __call__(self, since, until):
        range_expression = {}
        if since:
            range_expression['gte'] = since.isoformat()

        if until:
            range_expression['lte'] = until.isoformat()

        server_modified_missing = {
            "missing": {
                "field": "server_modified_on",
                "null_value": True,
                "existence": True
            }
        }
        query = json.dumps({
            'filter': {
                "or": [{
                    "and": [{
                        "not": server_modified_missing
                    }, {
                        "range": {
                            "server_modified_on": range_expression
                        }
                    }]
                },
                       {
                           "and": [
                               server_modified_missing, {
                                   "range": {
                                       "received_on": range_expression
                                   }
                               }
                           ]
                       }]
            }
        })

        return {'_search': query}


DATE_PARAMS = {
    'indexed_on':
        SimpleSinceParams('indexed_on_start', 'indexed_on_end'),
    'server_date_modified':
        SimpleSinceParams(
            'server_date_modified_start', 'server_date_modified_end'
        ),  # used by messaging-events
    'date_last_activity':
        SimpleSinceParams('date_last_activity.gte', 'date_last_activity.lt'),
}


def get_paginator(
    resource,
    page_size=None,
    pagination_mode=PaginationMode.date_indexed,
):
    return {
        PaginationMode.date_indexed: {
            'form': DatePaginator('indexed_on', page_size),
            'case': DatePaginator('indexed_on', page_size),
            'messaging-event': DatePaginator('date_last_activity', page_size),
        },
        PaginationMode.date_modified: {
            'form':
                DatePaginator(
                    ['server_modified_on', 'received_on'],
                    page_size,
                    params=FormFilterSinceParams(),
                ),
            'case':
                DatePaginator('server_date_modified', page_size),
            'messaging-event':
                DatePaginator('date_last_activity', page_size),
        },
        PaginationMode.cursor: {
            'ucr': UCRPaginator(page_size),
        },
    }[pagination_mode].get(resource, SimplePaginator(page_size))


class CommCareHqEnv(DictEnv):
    """
    An environment providing primitives for pulling from the
    CommCareHq API.
    """

    def __init__(self, commcare_hq_client, page_size=None, until=None):
        self.commcare_hq_client = commcare_hq_client
        self.until = until
        self.page_size = page_size
        super(CommCareHqEnv, self).__init__({'api_data': self.api_data})

    @unwrap('checkpoint_manager')
    def api_data(
        self,
        resource,
        checkpoint_manager,
        payload=None,
        include_referenced_items=None
    ):
        if resource not in SUPPORTED_RESOURCES:
            raise ValueError('Unknown API resource "%s' % resource)

        paginator = get_paginator(
            resource, self.page_size, checkpoint_manager.pagination_mode
        )
        paginator.init(payload, include_referenced_items, self.until)
        initial_params = paginator.next_page_params_since(
            checkpoint_manager.since_param
        )
        return self.commcare_hq_client.iterate(
            resource,
            paginator,
            params=initial_params,
            checkpoint_manager=checkpoint_manager
        )

    def bind(self, name, value):
        raise CannotBind()

    def replace(self, data):
        raise CannotReplace()


class SimplePaginator(object):
    """
    Paginate based on the 'next' URL provided in the API response.
    """

    def __init__(self, page_size=None, params=None):
        page_size = page_size if page_size else 1000
        self.page_size = page_size
        self.params = params

    def init(self, payload=None, include_referenced_items=None, until=None):
        self.payload = dict(payload or {})  # Do not mutate passed-in dicts
        self.include_referenced_items = include_referenced_items
        self.until = until

    def next_page_params_since(self, since=None):
        params = self.payload
        params['limit'] = self.page_size

        if (since or self.until) and self.params:
            params.update(self.params(since, self.until))

        if self.include_referenced_items:
            params.update([
                (f'{referenced_item}__full', 'true')
                for referenced_item in self.include_referenced_items
            ])

        return params

    def next_page_params_from_batch(self, batch):
        if batch['meta']['next']:
            return parse_qs(urlparse(batch['meta']['next']).query)

    def set_checkpoint(self, *args, **kwargs):
        pass


class DatePaginator(SimplePaginator):
    """
    This paginator is designed to get around the issue of deep paging
    where the deeper the page the longer the query takes.

    Paginate records according to a date in the record. The params for
    the next batch will include a filter for the date of the last record
    in the previous batch.

    This also adds an ordering parameter to ensure that the records are
    ordered by the date field in ascending order.

    :param since_field: The name of the date field to use for pagination.
    :param page_size: Number of results to request in each page
    """

    DEFAULT_PARAMS = object()

    def __init__(self, since_field, page_size=None, params=DEFAULT_PARAMS):
        page_size = page_size if page_size else DEFAULT_PAGE_SIZE
        params = DATE_PARAMS[
            since_field] if params is DatePaginator.DEFAULT_PARAMS else params
        super(DatePaginator, self).__init__(page_size, params)
        self.since_field = since_field

    def next_page_params_since(self, since=None):
        params = super(DatePaginator, self).next_page_params_since(since)
        params['order_by'] = self.since_field
        return params

    def next_page_params_from_batch(self, batch):
        since_date = self.get_since_date(batch)
        if since_date:
            return self.next_page_params_since(since_date)

    def get_since_date(self, batch):
        try:
            last_obj = batch['objects'][-1]
        except IndexError:
            return

        if last_obj:
            if isinstance(self.since_field, list):
                for field in self.since_field:
                    since = last_obj.get(field)
                    if since:
                        break
            else:
                since = last_obj.get(self.since_field)

            if since:
                try:
                    return parse(
                        since,
                        # ignoretz since we assume utc, and use naive
                        # datetimes everywhere
                        ignoretz=True
                    )
                except ParserError:
                    return None

    def set_checkpoint(self, checkpoint_manager, batch, is_final):
        since_date = self.get_since_date(batch)
        if since_date:
            try:
                last_obj = batch['objects'][-1]
            except IndexError:
                last_obj = {}
            checkpoint_manager.set_checkpoint(
                since_date, is_final, doc_id=last_obj.get("id", None)
            )
        else:
            logger.warning(
                'Failed to get a checkpoint date from a batch of data.'
            )


class UCRPaginator(SimplePaginator):
    def __init__(self, page_size=None, *args, **kwargs):
        super().__init__(page_size, *args, **kwargs)
        self.page_size = page_size if page_size else DEFAULT_UCR_PAGE_SIZE

    def next_page_params_from_batch(self, batch):
        params = super(UCRPaginator, self).next_page_params_from_batch(batch)
        if params:
            return self.payload | params

    def next_page_params_since(self, since=None):
        params = self.payload
        params['cursor'] = since
        params["limit"] = self.page_size
        return params

    def set_checkpoint(self, checkpoint_manager, batch, is_final):
        cursor = self.next_page_params_from_batch(batch)['cursor'][0]
        checkpoint_manager.set_checkpoint(
            datetime.utcnow(),
            is_final,
            cursor=cursor,
        )



import hashlib
import struct
import tempfile
import unittest

from commcare_export import misc


class TestDigestFile(unittest.TestCase):

    def check_digest(self, contents):
        with tempfile.NamedTemporaryFile(
            prefix='commcare-export-test-', mode='wb'
        ) as file:
            file.write(contents)
            file.flush()
            file_digest = misc.digest_file(file.name)

        # Make sure the chunking does not mess with stuff
        assert file_digest == hashlib.md5(contents).hexdigest()

    def test_digest_file_ascii(self):
        self.check_digest('Hello'.encode('utf-8'))

    def test_digest_file_long(self):
        self.check_digest(('Hello' * 100000).encode('utf-8'))

    def test_digest_file_utf8(self):
        self.check_digest('Miércoles'.encode('utf-8'))

    def test_digest_file_binary(self):
        self.check_digest(struct.pack('III'.encode('ascii'), 1, 2, 3))



import argparse
import inspect
import logging
import sys

from commcare_export.cli import CLI_ARGS
from commcare_export.utils import confirm, get_checkpoint_manager, print_runs

EXIT_STATUS_ERROR = 1

logger = logging.getLogger(__name__)


class BaseCommand(object):
    slug = None
    help = None

    @classmethod
    def add_arguments(cls, parser):
        raise NotImplementedError

    def run(self, args):
        raise NotImplementedError


class ListHistoryCommand(BaseCommand):
    slug = 'history'
    help = """List export history. History will be filtered by arguments provided.

    This command only applies when exporting to a SQL database. The command
    lists the checkpoints that have been created by the command.
    """

    @classmethod
    def add_arguments(cls, parser):
        parser.add_argument(
            '--limit',
            default=10,
            help="Limit the number of export runs to display"
        )
        parser.add_argument('--output', required=True, help='SQL Database URL')
        shared_args = {'project', 'query', 'checkpoint_key', 'commcare_hq'}
        for arg in CLI_ARGS:
            if arg.name in shared_args:
                arg.add_to_parser(parser)

    def run(self, args):
        manager = get_checkpoint_manager(args, require_query=False)
        manager.create_checkpoint_table()

        print("Listing checkpoints (most recent {}):".format(args.limit))
        if args.project:
            print("    project:        {}".format(args.project))
        if args.commcare_hq != 'prod':
            print("    commcare-hq:    {}".format(args.commcare_hq))
        if args.query:
            print("    query filename: {}".format(args.query))
        if manager.key:
            print("    key:            {}".format(manager.key))

        runs = manager.list_checkpoints(args.limit)
        print_runs(runs)


class SetKeyCommand(BaseCommand):
    slug = 'set-checkpoint-key'
    help = """Set the key for a particular checkpoint.

    This command is used to migrate an non-keyed checkpoint to a keyed
    checkpoint.

    This is useful if you already have a populated export database and do
    not wish to trigger rebuilds after editing the query file.

    For example, you've been running the export tool with query file A.xlsx
    and have a fully populated database. Now you need to add an extra column
    to the table but only want to populate it with new data.

    What you need to do is update your current checkpoint with a key that
    you can then use when running the command from now on.

        $ commcare-export-utils set-key --project X --query A.xlsx \\
            --output [SQL URL] --checkpoint-key my-key

    Now when you run the export tool in future you can use this key:

        $ commcare-export --project X --query A.xlsx --output [SQL URL] \\
            --checkpoint-key my-key ...

    """

    @classmethod
    def add_arguments(cls, parser):
        parser.add_argument('--output', required=True, help='SQL Database URL')
        shared_args = {'project', 'query', 'checkpoint_key'}
        for arg in CLI_ARGS:
            if arg.name in shared_args:
                arg.add_to_parser(parser, required=True)
            elif arg.name == 'commcare_hq':
                arg.add_to_parser(parser)

    def run(self, args):
        key = args.checkpoint_key
        manager = get_checkpoint_manager(args)
        manager.create_checkpoint_table()
        run_with_key = manager.list_checkpoints(limit=1)

        if run_with_key:
            print("A checkpoint with that key already exists.")
            return

        manager.key = None
        runs_no_key = manager.get_latest_checkpoints()

        if not runs_no_key:
            print(args)
            print("No checkpoint found with args matching those provided.")
            return

        print_runs(runs_no_key)
        if confirm(
            f"Do you want to set the key for this checkpoint to '{key}'"
        ):
            for checkpoint in runs_no_key:
                checkpoint.key = key
                manager.update_checkpoint(checkpoint)

        print("\nUpdated checkpoint:")
        print_runs(runs_no_key)


COMMANDS = [ListHistoryCommand, SetKeyCommand]


def main(argv):
    parser = argparse.ArgumentParser('commcare-export-utils')
    subparsers = parser.add_subparsers(dest='command')
    for command_type in COMMANDS:
        sub = subparsers.add_parser(
            command_type.slug,
            help=inspect.cleandoc(command_type.help).splitlines()[0],
            description=inspect.cleandoc(command_type.help),
            formatter_class=argparse.RawDescriptionHelpFormatter
        )
        command_type.add_arguments(sub)

    args = parser.parse_args(argv)

    logging.basicConfig(
        level=logging.WARN,
        format='%(asctime)s %(name)-12s %(levelname)-8s %(message)s'
    )

    exit(main_with_args(args))


def main_with_args(args):
    command = [c for c in COMMANDS if c.slug == args.command][0]
    command().run(args)


def entry_point():
    main(sys.argv[1:])


if __name__ == '__main__':
    entry_point()



import unittest

from commcare_export.checkpoint import CheckpointManagerWithDetails
from commcare_export.commcare_minilinq import (
    DEFAULT_UCR_PAGE_SIZE,
    PaginationMode,
    get_paginator,
)


class PaginatorTest(unittest.TestCase):
    def test_ucr_paginator_page_size(self):
        checkpoint_manager = CheckpointManagerWithDetails(
            None, None, PaginationMode.cursor
        )
        paginator = get_paginator(
            resource="ucr",
            pagination_mode=checkpoint_manager.pagination_mode)
        paginator.init()
        initial_params = paginator.next_page_params_since(
            checkpoint_manager.since_param
        )
        self.assertEqual(initial_params["limit"], DEFAULT_UCR_PAGE_SIZE)

        paginator = get_paginator(
            resource="ucr",
            page_size=1,
            pagination_mode=checkpoint_manager.pagination_mode)
        paginator.init()
        initial_params = paginator.next_page_params_since(
            checkpoint_manager.since_param
        )
        self.assertEqual(initial_params["limit"], 1)


import sys

from commcare_export import misc
from commcare_export.checkpoint import CheckpointManager
from commcare_export.specs import TableSpec
from commcare_export.writers import StreamingMarkdownTableWriter


def get_checkpoint_manager(args, require_query=True):
    md5 = None
    try:
        md5 = misc.digest_file(args.query)
    except Exception:
        if require_query:
            raise

    return CheckpointManager(
        args.output,
        args.query,
        md5,
        args.project,
        args.commcare_hq,
        args.checkpoint_key
    )


def confirm(message):
    confirm = input(f"""
        {message}? [y/N]
        """)
    return confirm == "y"


def print_runs(runs):
    print()
    rows = []
    for run in runs:
        rows.append([
            run.time_of_run,
            run.since_param,
            "True" if run.final else "False",
            run.project,
            run.query_file_name,
            run.query_file_md5,
            run.key,
            run.table_name,
            run.commcare
        ])

    rows = [[val if val is not None else '' for val in row] for row in rows]

    StreamingMarkdownTableWriter(
        sys.stdout, compute_widths=True
    ).write_table(TableSpec(
        name='',
        headings=[
            "Checkpoint Time",
            "Batch end date",
            "Export Complete",
            "Project",
            "Query Filename",
            "Query MD5",
            "Key",
            "Table",
            "CommCare HQ"
        ],
        rows=rows,
    ))



import unittest

from commcare_export.map_format import parse_function_arg, parse_template
from commcare_export.minilinq import Apply, Literal, Reference


class TestMapFormats(unittest.TestCase):

    def test_parse_template_no_args(self):
        expected = Apply(
            Reference('template'), Literal('my name is {}'),
            Reference('form.question1')
        )
        assert parse_template(
            Reference('form.question1'), 'template(my name is {})'
        ) == expected

    def test_parse_template_args(self):
        expected = Apply(
            Reference('template'), Literal('my name is {}'),
            Reference('form.question2')
        )
        assert parse_template(
            'form.question1', 'template(my name is {}, form.question2)'
        ) == expected

    def test_parse_template_args_long(self):
        expected = Apply(
            Reference('template'),
            Literal('https://www.commcarehq.org/a/{}/reports/form_data/{}/'),
            Reference('$.domain'),
            Reference('$.id'),
        )
        assert parse_template(
            'form.id',

            'template(https://www.commcarehq.org/a/{}/reports/form_data/{}/, '
            '$.domain, $.id)'
        ) == expected

    def test_parse_template_no_template(self):
        expected = Literal(
            'Error: template function requires the format template: template()'
        )
        assert parse_template('form.question1', 'template()') == expected

    def test_parse_function_arg_with_brackets(self):
        value_returned = parse_function_arg(
            'selected', 'selected(Other_(Specify))'
        )
        assert value_returned == 'Other_(Specify)'

    def test_parse_function_arg_empty_returns(self):
        value_returned = parse_function_arg('selected', 'selected()')
        assert value_returned == ''



import types
import unittest
from datetime import datetime
from itertools import *

import pytest
from commcare_export.env import *
from commcare_export.excel_query import get_value_or_root_expression
from commcare_export.minilinq import *
from commcare_export.writers import JValueTableWriter


class LazinessException(Exception):
    pass


def die(msg):
    # Hack: since "raise" is a statement not an expression, need a
    # funcall wrapping it
    raise LazinessException(msg)


class TestMiniLinq(unittest.TestCase):

    @classmethod
    def setup_class(cls):
        pass

    def check_case(self, val, expected):
        if isinstance(expected, list):
            assert [unwrap_val(datum) for datum in val] == expected
        else:
            assert val == expected

    def test_eval_literal(self):
        env = BuiltInEnv()
        assert Literal("foo").eval(env) == "foo"
        assert Literal(2).eval(env) == 2
        assert Literal({'foo': 'baz'}).eval(env) == {'foo': 'baz'}

    def test_eval_reference(self):
        env = BuiltInEnv()
        assert Reference("foo").eval(DictEnv({'foo': 2})) == 2
        assert Reference(Reference(Reference('a'))
                        ).eval(DictEnv({
                            'a': 'b',
                            'b': 'c',
                            'c': 2
                        })) == 2
        self.check_case(
            Reference("foo[*]").eval(JsonPathEnv({'foo': [2]})), [2]
        )
        # Should work the same w/ iterators as with lists
        self.check_case(
            Reference("foo[*]").eval(JsonPathEnv({'foo': range(0, 1)})), [0]
        )

        # Should be able to get back out to the root, as the JsonPathEnv
        # actually passes the full datum around
        self.check_case(
            Reference("foo.$.baz").eval(JsonPathEnv({
                'foo': [2],
                'baz': 3
            })), [3]
        )

    def test_eval_auto_id_reference(self):
        """
        Test that we have turned on the jsonpath_ng.jsonpath.auto_id
        field properly
        """
        env = BuiltInEnv()

        self.check_case(
            Reference("foo.id").eval(JsonPathEnv({'foo': [2]})), ['foo']
        )

        # When auto id is on, this always becomes a string. Sorry!
        self.check_case(
            Reference("foo.id").eval(JsonPathEnv({'foo': {
                'id': 2
            }})), ['2']
        )

    def test_eval_auto_id_reference_nested(self):
        # this test is documentation of existing (weird) functionality
        # that results from a combination of jsonpath_ng auto_id feature
        # and JsonPathEnv.lookup (which adds an additional auto ID for
        # some reason).
        env = JsonPathEnv({})

        flatmap = FlatMap(
            source=Literal([{
                "id": 1,
                "foo": {
                    'id': 'bid',
                    'name': 'bob'
                },
                "bar": [{
                    'baz': 'a1'
                }, {
                    'baz': 'a2',
                    'id': 'bazzer'
                }]
            }]),
            body=Reference('bar.[*]')
        )
        mmap = Map(
            source=flatmap,
            body=List([
                Reference("id"),
                Reference('baz'),
                Reference('$.id'),
                Reference('$.foo.id'),
                Reference('$.foo.name')
            ])
        )
        self.check_case(
            mmap.eval(env), [["1.bar.'1.bar.[0]'", 'a1', '1', '1.bid', 'bob'],
                             ['1.bar.bazzer', 'a2', '1', '1.bid', 'bob']]
        )

        # Without the additional auto id field added in JsonPathEnv the
        # result for Reference("id") changes as follows:
        #   '1.bar.1.bar.[0]' -> '1.bar.[0]'

        # With the change above AND a change to jsonpath_ng to prevent
        # converting IDs that exist into auto IDs (see
        # https://github.com/kennknowles/python-jsonpath-rw/pull/96) we
        # get the following:
        #   Reference("id"):
        #       '1.bar.bazzer' -> 'bazzer'
        #
        #   Reference('$.foo.id'):
        #       '1.bid' -> 'bid'

    def test_value_or_root(self):
        """
        Test that when accessing a child object the child data is used
        if it exists (normal case).
        """
        data = {"id": 1, "bar": [{'baz': 'a1'}, {'baz': 'a2'}]}
        self._test_value_or_root([Reference('id'),
                                  Reference('baz')], data, [
                                      ["1.bar.'1.bar.[0]'", 'a1'],
                                      ["1.bar.'1.bar.[1]'", 'a2'],
                                  ])

    def test_value_or_root_empty_list(self):
        """Should use the root object if the child is an empty list"""
        data = {
            "id": 1,
            "foo": "I am foo",
            "bar": [],
        }
        self._test_value_or_root([
            Reference('id'),
            Reference('baz'),
            Reference('$.foo')
        ], data, [
            ['1', [], "I am foo"],
        ])

    def test_value_or_root_empty_dict(self):
        """Should use the root object if the child is an empty dict"""
        data = {
            "id": 1,
            "foo": "I am foo",
            "bar": {},
        }
        self._test_value_or_root([
            Reference('id'),
            Reference('baz'),
            Reference('$.foo')
        ], data, [
            ['1', [], "I am foo"],
        ])

    def test_value_or_root_None(self):
        """Should use the root object if the child is None"""
        data = {
            "id": 1,
            "bar": None,
        }
        self._test_value_or_root([Reference('id'),
                                  Reference('baz')], data, [
                                      ['1', []],
                                  ])

    def test_value_or_root_missing(self):
        """Should use the root object if the child does not exist"""
        data = {
            "id": 1,
            "foo": "I am foo",
            # 'bar' is missing
        }
        self._test_value_or_root([
            Reference('id'),
            Reference('baz'),
            Reference('$.foo')
        ], data, [
            ['1', [], 'I am foo'],
        ])

    def test_value_or_root_ignore_field_in_root(self):
        """
        Test that a child reference is ignored if we are using the root
        doc even if there is a field with that name. (this doesn't apply
        to 'id')
        """
        data = {
            "id": 1,
            "foo": "I am foo",
        }
        self._test_value_or_root([Reference('id'),
                                  Reference('foo')], data, [
                                      ['1', []],
                                  ])

    def _test_value_or_root(self, columns, data, expected):
        """Low level test case for 'value-or-root'"""
        env = BuiltInEnv() | JsonPathEnv({})
        value_or_root = get_value_or_root_expression('bar.[*]')
        flatmap = FlatMap(source=Literal([data]), body=value_or_root)
        mmap = Map(source=flatmap, body=List(columns))
        self.check_case(mmap.eval(env), expected)

    def test_eval_collapsed_list(self):
        """
        Special case to handle XML -> JSON conversion where there just
        happened to be a single value at save time
        """
        env = BuiltInEnv()
        self.check_case(Reference("foo[*]").eval(JsonPathEnv({'foo': 2})), [2])
        assert Apply(Reference("*"), Literal(2), Literal(3)).eval(env) == 6
        assert Apply(Reference(">"), Literal(56),
                     Literal(23.5)).eval(env) == True
        assert Apply(Reference("len"), Literal([1, 2, 3])).eval(env) == 3
        assert Apply(Reference("bool"), Literal('a')).eval(env) == True
        assert Apply(Reference("bool"), Literal('')).eval(env) == False
        assert Apply(Reference("str2bool"), Literal('true')).eval(env) == True
        assert Apply(Reference("str2bool"), Literal('t')).eval(env) == True
        assert Apply(Reference("str2bool"), Literal('1')).eval(env) == True
        assert Apply(Reference("str2bool"), Literal('0')).eval(env) == False
        assert Apply(Reference("str2bool"),
                     Literal('false')).eval(env) == False
        assert Apply(Reference("str2bool"), Literal(u'日本')).eval(env) == False
        assert Apply(Reference("str2num"), Literal('10')).eval(env) == 10
        assert Apply(Reference("str2num"), Literal('10.56')).eval(env) == 10.56
        assert Apply(Reference("str2num"), Literal('')).eval(env) == None
        assert Apply(Reference("str2date"),
                     Literal('2015-01-01')).eval(env) == datetime(2015, 1, 1)
        assert Apply(Reference("str2date"), Literal('2015-01-01T18:32:57')
                    ).eval(env) == datetime(2015, 1, 1, 18, 32, 57)
        assert Apply(
            Reference("str2date"), Literal('2015-01-01T18:32:57.001200')
        ).eval(env) == datetime(2015, 1, 1, 18, 32, 57)
        assert Apply(
            Reference("str2date"), Literal('2015-01-01T18:32:57.001200Z')
        ).eval(env) == datetime(2015, 1, 1, 18, 32, 57)
        assert Apply(Reference("str2date"),
                     Literal(u'日'.encode('utf8'))).eval(env) == None
        assert Apply(Reference("str2date"), Literal(u'日')).eval(env) == None
        assert Apply(Reference("format-uuid"),
                     Literal(0xf00)).eval(env) == None
        assert Apply(Reference("format-uuid"),
                     Literal('f00')).eval(env) == None
        assert Apply(
            Reference("format-uuid"),
            Literal('00a3e019-4ce1-4587-94c5-0971dee2de22')
        ).eval(env) == '00a3e019-4ce1-4587-94c5-0971dee2de22'
        assert Apply(Reference("selected-at"), Literal('a b c'),
                     Literal('1')).eval(env) == 'b'
        assert Apply(
            Reference("selected-at"), Literal(u'a b 日'), Literal('-1')
        ).eval(env) == u'日'
        assert Apply(Reference("selected-at"), Literal('a b c'),
                     Literal('5')).eval(env) is None
        assert Apply(Reference("selected"), Literal('a b c'),
                     Literal('b')).eval(env) is True
        assert Apply(Reference("selected"), Literal(u'a b 日本'),
                     Literal('d')).eval(env) is False
        assert Apply(Reference("selected"), Literal(u'a bb 日本'),
                     Literal('b')).eval(env) is False
        assert Apply(
            Reference("selected"), Literal(u'a bb 日本'), Literal(u'日本')
        ).eval(env) is True
        assert Apply(
            Reference("join"), Literal('.'), Literal('a'), Literal('b'),
            Literal('c')
        ).eval(env) == 'a.b.c'
        assert Apply(Reference("default"), Literal(None),
                     Literal('a')).eval(env) == 'a'
        assert Apply(Reference("default"), Literal('b'),
                     Literal('a')).eval(env) == 'b'
        assert Apply(Reference("count-selected"),
                     Literal(u'a bb 日本')).eval(env) == 3
        assert Apply(Reference("sha1"), Literal(u'a bb 日本')
                    ).eval(env) == 'e25a54025417b06d88d40baa8c71f6eee9c07fb1'
        assert Apply(Reference("sha1"), Literal(b'2015')
                    ).eval(env) == '9cdda67ded3f25811728276cefa76b80913b4c54'
        assert Apply(Reference("sha1"), Literal(2015)
                    ).eval(env) == '9cdda67ded3f25811728276cefa76b80913b4c54'

    def test_or(self):
        env = BuiltInEnv()
        assert Apply(Reference("or"), Literal(None), Literal(2)).eval(env) == 2

        laziness_iterator = RepeatableIterator(
            lambda: (i if i < 1 else die('Not lazy enough') for i in range(2))
        )
        assert Apply(Reference("or"), Literal(1),
                     Literal(laziness_iterator)).eval(env) == 1
        assert Apply(Reference("or"), Literal(''),
                     Literal(laziness_iterator)).eval(env) == ''
        assert Apply(Reference("or"), Literal(0),
                     Literal(laziness_iterator)).eval(env) == 0
        with pytest.raises(LazinessException):
            Apply(Reference("or"), Literal(None),
                  Literal(laziness_iterator)).eval(env)

        env = env | JsonPathEnv({'a': {'c': 'c val'}})
        assert Apply(Reference("or"), Reference('a.b'),
                     Reference('a.c')).eval(env) == 'c val'
        assert Apply(Reference("or"), Reference('a.b'),
                     Reference('a.d')).eval(env) is None

        env = env.replace({'a': [], 'b': [1, 2], 'c': 2})
        self.check_case(
            Apply(Reference("or"), Reference('a.[*]'),
                  Reference('b')).eval(env), [1, 2]
        )
        self.check_case(
            Apply(Reference("or"), Reference('b.[*]'),
                  Reference('c')).eval(env), [1, 2]
        )
        self.check_case(
            Apply(Reference("or"), Reference('a.[*]'),
                  Reference('$')).eval(env), {
                      'a': [],
                      'b': [1, 2],
                      'c': 2,
                      'id': '$'
                  }
        )

    def test_attachment_url(self):
        env = BuiltInEnv({'commcarehq_base_url': 'https://www.commcarehq.org'}
                        ) | JsonPathEnv({
                            'id': '123',
                            'domain': 'd1',
                            'photo': 'a.jpg'
                        })
        expected = 'https://www.commcarehq.org/a/d1/api/form/attachment/123/a.jpg'
        assert Apply(Reference('attachment_url'),
                     Reference('photo')).eval(env) == expected

    def test_attachment_url_repeat(self):
        env = BuiltInEnv({'commcarehq_base_url': 'https://www.commcarehq.org'}
                        ) | JsonPathEnv({
                            'id': '123',
                            'domain': 'd1',
                            'repeat': [{
                                'photo': 'a.jpg'
                            }, {
                                'photo': 'b.jpg'
                            }]
                        })
        expected = [
            'https://www.commcarehq.org/a/d1/api/form/attachment/123/a.jpg',
            'https://www.commcarehq.org/a/d1/api/form/attachment/123/b.jpg',
        ]
        result = unwrap_val(
            Map(
                source=Reference('repeat.[*]'),
                body=Apply(Reference('attachment_url'), Reference('photo'))
            ).eval(env)
        )
        assert result == expected

    def test_form_url(self):
        env = BuiltInEnv({'commcarehq_base_url': 'https://www.commcarehq.org'}
                        ) | JsonPathEnv({
                            'id': '123',
                            'domain': 'd1'
                        })
        expected = 'https://www.commcarehq.org/a/d1/reports/form_data/123/'
        assert Apply(Reference('form_url'),
                     Reference('id')).eval(env) == expected

    def test_case_url(self):
        env = BuiltInEnv({'commcarehq_base_url': 'https://www.commcarehq.org'}
                        ) | JsonPathEnv({
                            'id': '123',
                            'domain': 'd1'
                        })
        expected = 'https://www.commcarehq.org/a/d1/reports/case_data/123/'
        assert Apply(Reference('case_url'),
                     Reference('id')).eval(env) == expected

    def test_unique(self):
        env = BuiltInEnv() | JsonPathEnv({
            "list": [{
                "a": 1
            }, {
                "a": 2
            }, {
                "a": 3
            }, {
                "a": 2
            }]
        })
        assert Apply(Reference('unique'),
                     Reference('list[*].a')).eval(env) == [1, 2, 3]

    def test_template(self):
        env = BuiltInEnv() | JsonPathEnv({'a': '1', 'b': '2'})
        assert Apply(
            Reference('template'), Literal('{}.{}'), Reference('a'),
            Reference('b')
        ).eval(env) == '1.2'

    def test_substr(self):
        env = BuiltInEnv({
            'single_byte_chars': u'abcdefghijklmnopqrstuvwxyz',
            'multi_byte_chars': u'αβγδεζηθικλμνξοπρςστυφχψω',
            'an_integer': 123456
        })
        assert Apply(
            Reference('substr'), Reference('single_byte_chars'), Literal(-4),
            Literal(30)
        ).eval(env) == None
        assert Apply(
            Reference('substr'), Reference('single_byte_chars'), Literal(0),
            Literal(26)
        ).eval(env) == u'abcdefghijklmnopqrstuvwxyz'
        assert Apply(
            Reference('substr'), Reference('single_byte_chars'), Literal(10),
            Literal(16)
        ).eval(env) == u'klmnop'
        assert Apply(
            Reference('substr'), Reference('single_byte_chars'), Literal(13),
            Literal(14)
        ).eval(env) == u'n'
        assert Apply(
            Reference('substr'), Reference('single_byte_chars'), Literal(13),
            Literal(13)
        ).eval(env) == u''
        assert Apply(
            Reference('substr'), Reference('single_byte_chars'), Literal(14),
            Literal(13)
        ).eval(env) == u''
        assert Apply(
            Reference('substr'), Reference('single_byte_chars'), Literal(5),
            Literal(-1)
        ).eval(env) == None

        assert Apply(
            Reference('substr'), Reference('multi_byte_chars'), Literal(-4),
            Literal(30)
        ).eval(env) == None
        assert Apply(
            Reference('substr'), Reference('multi_byte_chars'), Literal(0),
            Literal(25)
        ).eval(env) == u'αβγδεζηθικλμνξοπρςστυφχψω'
        assert Apply(
            Reference('substr'), Reference('multi_byte_chars'), Literal(10),
            Literal(15)
        ).eval(env) == u'λμνξο'
        assert Apply(
            Reference('substr'), Reference('multi_byte_chars'), Literal(13),
            Literal(14)
        ).eval(env) == u'ξ'
        assert Apply(
            Reference('substr'), Reference('multi_byte_chars'), Literal(13),
            Literal(12)
        ).eval(env) == u''
        assert Apply(
            Reference('substr'), Reference('multi_byte_chars'), Literal(14),
            Literal(13)
        ).eval(env) == u''
        assert Apply(
            Reference('substr'), Reference('multi_byte_chars'), Literal(5),
            Literal(-1)
        ).eval(env) == None

        assert Apply(
            Reference('substr'), Reference('an_integer'), Literal(-1),
            Literal(3)
        ).eval(env) == None
        assert Apply(
            Reference('substr'), Reference('an_integer'), Literal(0),
            Literal(6)
        ).eval(env) == u'123456'
        assert Apply(
            Reference('substr'), Reference('an_integer'), Literal(2),
            Literal(4)
        ).eval(env) == u'34'
        assert Apply(
            Reference('substr'), Reference('an_integer'), Literal(4),
            Literal(2)
        ).eval(env) == u''
        assert Apply(
            Reference('substr'), Reference('an_integer'), Literal(5),
            Literal(-1)
        ).eval(env) == None

    def test_map(self):
        env = BuiltInEnv() | DictEnv({})

        laziness_iterator = RepeatableIterator(
            lambda: ({
                'a': i
            } if i < 5 else die('Not lazy enough') for i in range(12))
        )

        assert list(
            Map(
                source=Literal([{
                    'a': 1
                }, {
                    'a': 2
                }, {
                    'a': 3
                }]),
                body=Literal(1)
            ).eval(env)
        ) == [1, 1, 1]
        assert list(
            Map(
                source=Literal([{
                    'a': 1
                }, {
                    'a': 2
                }, {
                    'a': 3
                }]),
                body=Reference('a')
            ).eval(env)
        ) == [1, 2, 3]

        assert list(
            islice(
                Map(source=Literal(laziness_iterator),
                    body=Reference('a')).eval(env), 5
            )
        ) == [0, 1, 2, 3, 4]

        try:
            list(
                Map(source=Literal(laziness_iterator),
                    body=Reference('a')).eval(env)
            )
            raise Exception('Should have failed')
        except LazinessException:
            pass

    def test_flatmap(self):
        env = BuiltInEnv() | DictEnv({})

        laziness_iterator = RepeatableIterator(
            lambda: ({
                'a': range(i)
            } if i < 4 else die('Not lazy enough') for i in range(12))
        )

        assert list(
            FlatMap(
                source=Literal([{
                    'a': [1]
                }, {
                    'a': 'foo'
                }, {
                    'a': [3, 4]
                }]),
                body=Literal([1, 2, 3])
            ).eval(env)
        ) == [1, 2, 3, 1, 2, 3, 1, 2, 3]
        assert list(
            FlatMap(
                source=Literal([{
                    'a': [1]
                }, {
                    'a': [2]
                }, {
                    'a': [3, 4]
                }]),
                body=Reference('a')
            ).eval(env)
        ) == [1, 2, 3, 4]

        assert list(
            islice(
                FlatMap(
                    source=Literal(laziness_iterator), body=Reference('a')
                ).eval(env), 6
            )
        ) == [0, 0, 1, 0, 1, 2]

        try:
            list(
                FlatMap(
                    source=Literal(laziness_iterator), body=Reference('a')
                ).eval(env)
            )
            raise Exception('Should have failed')
        except LazinessException:
            pass

    def _setup_emit_test(self, emitter_env):
        env = BuiltInEnv() | JsonPathEnv({
            'foo': {
                'baz': 3,
                'bar': True,
                'boo': None
            }
        }) | emitter_env
        Emit(
            table='Foo',
            headings=[Literal('foo')],
            source=List([
                List([
                    Reference('foo.baz'),
                    Reference('foo.bar'),
                    Reference('foo.foo'),
                    Reference('foo.boo')
                ])
            ]),
            missing_value='---'
        ).eval(env)

    def test_emit(self):
        writer = JValueTableWriter()
        self._setup_emit_test(EmitterEnv(writer))
        assert list(writer.tables['Foo'].rows) == [[3, True, '---', None]]

    def test_emit_generator(self):

        class TestWriter(JValueTableWriter):

            def write_table(self, table):
                self.tables[table.name] = table

        writer = TestWriter()
        self._setup_emit_test(EmitterEnv(writer))
        assert isinstance(
            writer.tables['Foo'].rows, (map, filter, types.GeneratorType)
        )

    def test_emit_env_generator(self):

        class TestEmitterEnv(EmitterEnv):

            def emit_table(self, table_spec):
                self.table = table_spec

        env = TestEmitterEnv(JValueTableWriter())
        self._setup_emit_test(env)
        assert isinstance(env.table.rows, (map, filter, types.GeneratorType))

    def test_emit_multi_same_query(self):
        """
        Test that we can emit multiple tables from the same set of
        source data. This is useful if you need to generate multiple
        tables from the same datasource.
        """
        writer = JValueTableWriter()
        env = BuiltInEnv() | JsonPathEnv() | EmitterEnv(writer)

        result = Map(
            source=Literal([
                {
                    'foo': {
                        'baz': 3,
                        'bar': True,
                        'boo': None
                    }
                },
                {
                    'foo': {
                        'baz': 4,
                        'bar': False,
                        'boo': 1
                    }
                },
            ]),
            body=List([
                Emit(
                    table='FooBaz',
                    headings=[Literal('foo')],
                    source=List([List([Reference('foo.baz')])]),
                ),
                Emit(
                    table='FooBar',
                    headings=[Literal('foo')],
                    source=List([List([Reference('foo.bar')])]),
                )
            ]),
        ).eval(env)

        # evaluate result
        list(result)
        assert 2 == len(writer.tables)
        assert writer.tables['FooBaz'].rows == [[3], [4]]
        assert writer.tables['FooBar'].rows == [[True], [False]]

    def test_emit_mutli_different_query(self):
        """
        Test that we can emit multiple tables from the same set of
        source data even if the emitted table have different 'root doc'
        expressions.

        Example use case could be emitting cases and case actions, or
        form data and repeats.
        """
        writer = JValueTableWriter()
        env = BuiltInEnv() | JsonPathEnv() | EmitterEnv(writer)
        result = Filter(  # the filter here is to prevent accumulating a `[None]` value for each doc
            predicate=Apply(
                Reference("filter_empty"),
                Reference("$")
            ),
            source=Map(
                # in practice `source` would be and api query such as
                # Apply(Reference('api_data'), Literal('case'), Literal({'type': 'case_type'}))
                source=Literal([
                    {'id': 1, 'foo': {'baz': 3, 'bar': True, 'boo': None}, 'actions': [{'a': 3}, {'a': 4}]},
                    {'id': 2, 'foo': {'baz': 4, 'bar': False, 'boo': 1}, 'actions': [{'a': 5}, {'a': 6}]},
                ]),
                body=List([
                    Emit(
                        table="t1",
                        headings=[Literal("id")],
                        source=Map(
                            source=Reference("`this`"),
                            body=List([
                                Reference("id"),
                            ]),
                        )
                    ),
                    Emit(
                        table="t2",
                        headings=[Literal("id"), Literal("a")],
                        source=Map(
                            source=Reference("actions[*]"),
                            body=List([
                                Reference("$.id"),
                                Reference("a")
                            ]),
                        )
                    )
                ])
            )
        ).eval(env)

        # evaluate result
        list(result)
        assert writer.tables['t1'].rows == [['1'], ['2']]
        assert writer.tables['t2'].rows == [['1', 3], ['1', 4], ['2', 5],
                                            ['2', 6]]

    def test_from_jvalue(self):
        assert MiniLinq.from_jvalue({"Ref": "form.log_subreport"}
                                   ) == Reference("form.log_subreport")
        assert (
            MiniLinq.from_jvalue({
                "Apply": {
                    "fn": {
                        "Ref": "len"
                    },
                    "args": [{
                        "Ref": "form.log_subreport"
                    }]
                }
            }) == Apply(Reference("len"), Reference("form.log_subreport"))
        )
        assert MiniLinq.from_jvalue([{
            "Ref": "form.log_subreport"
        }]) == [Reference("form.log_subreport")]

    def test_filter(self):
        env = BuiltInEnv() | DictEnv({})
        named = [{'n': n} for n in range(1, 5)]
        assert list(
            Filter(
                Literal(named),
                Apply(Reference('>'), Reference('n'), Literal(2))
            ).eval(env)
        ) == [{
            'n': 3
        }, {
            'n': 4
        }]
        assert list(
            Filter(
                Literal([1, 2, 3, 4]),
                Apply(Reference('>'), Reference('n'), Literal(2)), 'n'
            ).eval(env)
        ) == [3, 4]

    def test_emit_table_unwrap_dicts(self):
        writer = JValueTableWriter()
        env = EmitterEnv(writer)
        env.emit_table(
            TableSpec(
                **{
                    'name':
                        't1',
                    'headings': ['a'],
                    'rows': [
                        ['hi'],
                        [{
                            '#text': 'test_text',
                            '@case_type': 'person',
                            '@relationship': 'child',
                            'id': 'nothing'
                        }],
                        [{
                            '@case_type': '',
                            '@relationship': 'child',
                            'id': 'some_id'
                        }],
                        [{
                            't': 123
                        }],
                    ]
                }
            )
        )

        writer.tables['t1'].rows = [['hi'], ['test_text'], [''], [{'t': 123}]]



import sqlalchemy

DATA_TYPE_TEXT = 'text'
DATA_TYPE_BOOLEAN = 'boolean'
DATA_TYPE_DATE = 'date'
DATA_TYPE_DATETIME = 'datetime'
DATA_TYPE_INTEGER = 'integer'
DATA_TYPE_JSON = 'json'

DATA_TYPES_TO_SQLALCHEMY_TYPES = {
    DATA_TYPE_TEXT: sqlalchemy.Text(),
    DATA_TYPE_BOOLEAN: sqlalchemy.Boolean(),
    DATA_TYPE_DATETIME: sqlalchemy.DateTime(),
    DATA_TYPE_DATE: sqlalchemy.Date(),
    DATA_TYPE_INTEGER: sqlalchemy.Integer(),
    DATA_TYPE_JSON: sqlalchemy.JSON(),
}


class UnknownDataType(Exception):
    pass


def get_sqlalchemy_type(data_type):
    if data_type not in DATA_TYPES_TO_SQLALCHEMY_TYPES:
        raise UnknownDataType(data_type)
    else:
        return DATA_TYPES_TO_SQLALCHEMY_TYPES[data_type]



import os.path
import pprint
import unittest

import openpyxl

from commcare_export.builtin_queries import ColumnEnforcer
from commcare_export.env import BuiltInEnv, JsonPathEnv
from commcare_export.excel_query import *
from commcare_export.excel_query import _get_safe_source_field


class TestExcelQuery(unittest.TestCase):

    @classmethod
    def setup_class(cls):
        pass

    def test_split_leftmost(self):
        assert split_leftmost(
            parse_jsonpath('foo')
        ) == (jsonpath.Fields('foo'), jsonpath.This())
        assert split_leftmost(
            parse_jsonpath('foo.baz')
        ) == (jsonpath.Fields('foo'), jsonpath.Fields('baz'))
        assert split_leftmost(parse_jsonpath('foo.baz.bar')) == (
            jsonpath.Fields('foo'),
            jsonpath.Fields('baz').child(jsonpath.Fields('bar'))
        )
        assert split_leftmost(
            parse_jsonpath('[*].baz')
        ) == (jsonpath.Slice(), jsonpath.Fields('baz'))
        assert split_leftmost(parse_jsonpath('foo[*].baz')) == (
            jsonpath.Fields('foo'),
            jsonpath.Slice().child(jsonpath.Fields('baz'))
        )

    def test_get_safe_source_field(self):
        assert _get_safe_source_field(
            'foo.bar.baz') == Reference('foo.bar.baz')
        assert _get_safe_source_field('foo[*].baz') == Reference('foo[*].baz')
        assert _get_safe_source_field(
            'foo..baz[*]') == Reference('foo..baz[*]')
        assert _get_safe_source_field('foo.#baz') == Reference('foo."#baz"')
        assert _get_safe_source_field(
            'foo.bar[*]..%baz') == Reference('foo.bar[*].."%baz"')
        assert _get_safe_source_field(
            'foo.bar:1.baz') == Reference('foo."bar:1".baz')

        try:
            assert _get_safe_source_field('foo.bar.')
            assert False, "Expected exception"
        except Exception:
            pass

    def test_compile_mappings(self):
        test_cases = [
            (
                'mappings.xlsx', {
                    'a': {
                        'w': 12,
                        'x': 13,
                        'y': 14,
                        'z': 15,
                        'q': 16,
                        'r': 17,
                    },
                    'b': {
                        'www': 'hello',
                        'xxx': 'goodbye',
                        'yyy': 'what is up',
                    },
                    'c': {
                        1: 'foo',
                        2: 'bar',
                        3: 'biz',
                        4: 'bizzle',
                    }
                }
            ),
        ]

        def flatten(dd):
            if isinstance(dd, defaultdict):
                return dict([(key, flatten(val)) for key, val in dd.items()])
            else:
                return dd

        for filename, mappings in test_cases:
            abs_path = os.path.join(os.path.dirname(__file__), filename)
            compiled = compile_mappings(
                openpyxl.load_workbook(abs_path)['Mappings']
            )
            # Print will be suppressed by pytest unless it fails
            if not (flatten(compiled) == mappings):
                print('In %s:' % filename)
                pprint.pprint(flatten(compiled))
                print('!=')
                pprint.pprint(mappings)
            assert flatten(compiled) == mappings

    def test_parse_sheet(self):

        test_cases = [
            (
                '001_JustDataSource.xlsx',
                SheetParts(
                    name='Forms',
                    headings=[],
                    source=Apply(
                        Reference("api_data"), Literal("form"),
                        Reference('checkpoint_manager')
                    ),
                    body=None,
                    data_source="form"
                ),
            ),
            # (
            #     '001a_JustDataSource_LibreOffice.xlsx',
            #     Emit(
            #         table='Forms',
            #         headings=[],
            #         source=Apply(Reference("api_data"), Literal("form"))
            #     )
            # ),
            (
                '002_DataSourceAndFilters.xlsx',
                SheetParts(
                    name='Forms',
                    headings=[],
                    source=Apply(
                        Reference("api_data"), Literal("form"),
                        Reference("checkpoint_manager"),
                        Literal({
                            'app_id': ['foobizzle'],
                            'type': ['intake'],
                        })
                    ),
                    body=None,
                    data_source="form"
                )
            ),
            (
                '003_DataSourceAndEmitColumns.xlsx',
                SheetParts(
                    name='Forms',
                    headings=[
                        Literal('Form Type'),
                        Literal('Fecha de Nacimiento'),
                        Literal('Sexo'),
                        Literal('Danger 0'),
                        Literal('Danger 1'),
                        Literal('Danger Fever'),
                        Literal('Danger error'),
                        Literal('Danger error'),
                        Literal('special'),
                        Literal('Danger substring 1'),
                        Literal('Danger substring 2'),
                        Literal('Danger substring error 3'),
                        Literal('Danger substring error 4'),
                        Literal('Danger substring error 5')
                    ],
                    source=Apply(
                        Reference("api_data"), Literal("form"),
                        Reference('checkpoint_manager')
                    ),
                    body=List([
                        Reference("type"),
                        Apply(
                            Reference("FormatDate"),
                            Reference("date_of_birth")
                        ),
                        Apply(Reference("sexo"), Reference("gender")),
                        Apply(
                            Reference("selected-at"), Reference("dangers"),
                            Literal(0)
                        ),
                        Apply(
                            Reference("selected-at"), Reference("dangers"),
                            Literal(1)
                        ),
                        Apply(
                            Reference("selected"), Reference("dangers"),
                            Literal('fever')
                        ),
                        Literal(
                            'Error: selected-at index must be an integer: '
                            'selected-at(abc)'
                        ),
                        Literal('Error: Unable to parse: selected(fever'),
                        Reference('path."#text"'),
                        Apply(
                            Reference("substr"), Reference("dangers"),
                            Literal(0), Literal(10)
                        ),
                        Apply(
                            Reference("substr"), Reference("dangers"),
                            Literal(4), Literal(3)
                        ),
                        Literal(
                            'Error: both substr arguments must be '
                            'non-negative integers: substr(a, b)'
                        ),
                        Literal(
                            'Error: both substr arguments must be '
                            'non-negative integers: substr(-1, 10)'
                        ),
                        Literal(
                            'Error: both substr arguments must be '
                            'non-negative integers: substr(3, -4)'
                        )
                    ]),
                    data_source="form"
                )
            ),
            (
                '005_DataSourcePath.xlsx',
                SheetParts(
                    name='Forms',
                    headings=[],
                    source=Apply(
                        Reference("api_data"), Literal("form"),
                        Reference('checkpoint_manager')
                    ),
                    body=None,
                    root_expr=Reference(
                        'form.delivery_information.child_questions.[*]'
                    ),
                    data_source="form"
                )
            ),
            (
                '006_IncludeReferencedItems.xlsx',
                SheetParts(
                    name='Forms',
                    headings=[],
                    source=Apply(
                        Reference("api_data"), Literal("form"),
                        Reference("checkpoint_manager"), Literal(None),
                        Literal(['foo', 'bar', 'bizzle'])
                    ),
                    body=None,
                    data_source="form"
                )
            ),
            (
                '010_JustDataSourceTableName.xlsx',
                SheetParts(
                    name='my_table',
                    headings=[],
                    source=Apply(
                        Reference("api_data"), Literal("form"),
                        Reference('checkpoint_manager')
                    ),
                    body=None,
                    data_source="form"
                )
            ),
        ]

        for filename, minilinq in test_cases:
            # This output will be captured by pytest and printed in case
            # of failure; helpful to isolate which test case
            print(f'Compiling sheet {filename}')
            abs_path = os.path.join(os.path.dirname(__file__), filename)
            compiled = parse_sheet(openpyxl.load_workbook(abs_path).active)
            # Print will be suppressed by pytest unless it fails
            if not (compiled == minilinq):
                print(f'In {filename}:')
                pprint.pprint(compiled)
                print('!=')
                pprint.pprint(minilinq)
            assert compiled == minilinq

    def test_parse_workbook(self):
        field_mappings = {'t1': 'Form 1', 't2': 'Form 2'}
        test_cases = [
            (
                '004_TwoDataSources.xlsx', [
                    SheetParts(
                        name='Forms',
                        headings=[],
                        source=Apply(
                            Reference("api_data"), Literal("form"),
                            Reference('checkpoint_manager')
                        ),
                        body=None,
                        data_source="form"
                    ),
                    SheetParts(
                        name='Cases',
                        headings=[],
                        source=Apply(
                            Reference("api_data"), Literal("case"),
                            Reference('checkpoint_manager')
                        ),
                        body=None,
                        data_source="case"
                    )
                ]
            ),
            (
                '007_Mappings.xlsx', [
                    SheetParts(
                        name='Forms',
                        headings=[Literal('Form Type')],
                        source=Apply(
                            Reference("api_data"), Literal("form"),
                            Reference('checkpoint_manager')
                        ),
                        body=List([
                            compile_mapped_field(
                                field_mappings, Reference("type")
                            )
                        ]),
                        data_source="form"
                    )
                ]
            ),
        ]

        for filename, minilinq in test_cases:
            # This output will be captured by pytest and printed in case
            # of failure; helpful to isolate which test case
            print(f'Compiling workbook {filename}')
            abs_path = os.path.join(os.path.dirname(__file__), filename)
            compiled = parse_workbook(openpyxl.load_workbook(abs_path))
            # Print will be suppressed by pytest unless it fails
            if not (compiled == minilinq):
                print('In %s:' % filename)
                pprint.pprint(compiled)
                print('!=')
                pprint.pprint(minilinq)
            assert compiled == minilinq

    def test_compile_mapped_field(self):
        env = BuiltInEnv() | JsonPathEnv({'foo': {'bar': 'a', 'baz': 'b'}})
        expression = compile_mapped_field({'a': 'mapped from a'},
                                          Reference('foo.bar'))
        assert expression.eval(env) == 'mapped from a'

        expression = compile_mapped_field({'a': 'mapped from a'},
                                          Reference('foo.baz'))
        assert list(expression.eval(env))[0].value == 'b'

        expression = compile_mapped_field({'a': 'mapped from a'},
                                          Reference('foo.boo'))
        assert list(expression.eval(env)) == []

    def test_get_queries_from_excel(self):
        minilinq = Bind(
            'checkpoint_manager',
            Apply(
                Reference('get_checkpoint_manager'), Literal("form"),
                Literal(["Forms"])
            ),
            Emit(
                table='Forms',
                missing_value='---',
                headings=[
                    Literal('Form Type'),
                    Literal('Fecha de Nacimiento'),
                    Literal('Sexo'),
                    Literal('Danger 0'),
                    Literal('Danger 1'),
                    Literal('Danger Fever'),
                    Literal('Danger error'),
                    Literal('Danger error'),
                    Literal('special'),
                    Literal('Danger substring 1'),
                    Literal('Danger substring 2'),
                    Literal('Danger substring error 3'),
                    Literal('Danger substring error 4'),
                    Literal('Danger substring error 5')
                ],
                source=Map(
                    source=Apply(
                        Reference("api_data"), Literal("form"),
                        Reference('checkpoint_manager')
                    ),
                    body=List([
                        Reference("type"),
                        Apply(
                            Reference("FormatDate"),
                            Reference("date_of_birth")
                        ),
                        Apply(Reference("sexo"), Reference("gender")),
                        Apply(
                            Reference("selected-at"), Reference("dangers"),
                            Literal(0)
                        ),
                        Apply(
                            Reference("selected-at"), Reference("dangers"),
                            Literal(1)
                        ),
                        Apply(
                            Reference("selected"), Reference("dangers"),
                            Literal('fever')
                        ),
                        Literal(
                            'Error: selected-at index must be an integer: '
                            'selected-at(abc)'
                        ),
                        Literal('Error: Unable to parse: selected(fever'),
                        Reference('path."#text"'),
                        Apply(
                            Reference("substr"), Reference("dangers"),
                            Literal(0), Literal(10)
                        ),
                        Apply(
                            Reference("substr"), Reference("dangers"),
                            Literal(4), Literal(3)
                        ),
                        Literal(
                            'Error: both substr arguments must be '
                            'non-negative integers: substr(a, b)'
                        ),
                        Literal(
                            'Error: both substr arguments must be '
                            'non-negative integers: substr(-1, 10)'
                        ),
                        Literal(
                            'Error: both substr arguments must be '
                            'non-negative integers: substr(3, -4)'
                        )
                    ])
                )
            )
        )

        self._compare_minilinq_to_compiled(
            minilinq, '003_DataSourceAndEmitColumns.xlsx'
        )

    def test_alternate_source_fields(self):
        minilinq = List([
            # First sheet uses a CSV column and also tests combining "Map Via"
            Bind(
                'checkpoint_manager',
                Apply(
                    Reference('get_checkpoint_manager'), Literal("form"),
                    Literal(["Forms"])
                ),
                Emit(
                    table='Forms',
                    missing_value='---',
                    headings=[
                        Literal('dob'),
                    ],
                    source=Map(
                        source=Apply(
                            Reference("api_data"), Literal("form"),
                            Reference('checkpoint_manager')
                        ),
                        body=List([
                            Apply(
                                Reference("str2date"),
                                Apply(
                                    Reference("or"), Reference("dob"),
                                    Reference("date_of_birth"),
                                    Reference("d_o_b")
                                )
                            ),
                        ])
                    )
                )
            ),

            # Second sheet uses multiple alternate source field columns (listed out of order)
            Bind(
                'checkpoint_manager',
                Apply(
                    Reference('get_checkpoint_manager'), Literal("form"),
                    Literal(["Forms1"])
                ),
                Emit(
                    table='Forms1',
                    missing_value='---',
                    headings=[
                        Literal('dob'),
                        Literal('Sex'),
                    ],
                    source=Map(
                        source=Apply(
                            Reference("api_data"), Literal("form"),
                            Reference('checkpoint_manager')
                        ),
                        body=List([
                            Reference("dob"),
                            Apply(
                                Reference("or"), Reference("gender"),
                                Reference("sex"), Reference("sex0")
                            )
                        ])
                    )
                )
            ),
        ])

        self._compare_minilinq_to_compiled(
            minilinq, '011_AlternateSourceFields.xlsx'
        )

    def test_columns_with_data_types(self):
        minilinq = Bind(
            'checkpoint_manager',
            Apply(
                Reference('get_checkpoint_manager'), Literal("form"),
                Literal(["Forms"])
            ),
            Emit(
                table='Forms',
                missing_value='---',
                headings=[
                    Literal('Name'),
                    Literal('Date of Birth'),
                    Literal('No Type Set'),
                    Literal('A Number'),
                    Literal('Bad Type'),
                ],
                source=Map(
                    source=Apply(
                        Reference("api_data"), Literal("form"),
                        Reference('checkpoint_manager')
                    ),
                    body=List([
                        Reference("name"),
                        Reference("date_of_birth"),
                        Reference("no_type_set"),
                        Reference("a_number"),
                        Reference("bad_type"),
                    ])
                ),
                data_types=[
                    Literal('text'),
                    Literal('date'),
                    Literal(None),
                    Literal('integer'),
                    Literal('bad_type'),
                ],
            ),
        )
        self._compare_minilinq_to_compiled(
            minilinq, '012_ColumnsWithTypes.xlsx'
        )

    def test_multi_emit(self):
        minilinq = List([
            Bind(
                "checkpoint_manager",
                Apply(
                    Reference('get_checkpoint_manager'), Literal("form"),
                    Literal(["Forms", "Cases"])
                ),
                Filter(
                    predicate=Apply(Reference("filter_empty"), Reference("$")),
                    source=Map(
                        source=Apply(
                            Reference("api_data"), Literal("form"),
                            Reference('checkpoint_manager')
                        ),
                        body=List([
                            Emit(
                                table="Forms",
                                headings=[Literal("id"),
                                          Literal("name")],
                                missing_value='---',
                                source=Map(
                                    source=Reference("`this`"),
                                    body=List([
                                        Reference("id"),
                                        Reference("form.name"),
                                    ]),
                                )
                            ),
                            Emit(
                                table="Cases",
                                headings=[Literal("case_id")],
                                missing_value='---',
                                source=Map(
                                    source=Reference("form..case"),
                                    body=List([
                                        Reference("@case_id"),
                                    ]),
                                )
                            )
                        ])
                    )
                )
            ),
            Bind(
                'checkpoint_manager',
                Apply(
                    Reference('get_checkpoint_manager'), Literal("case"),
                    Literal(["Other cases"])
                ),
                Emit(
                    table="Other cases",
                    headings=[Literal("id")],
                    missing_value='---',
                    source=Map(
                        source=Apply(
                            Reference("api_data"), Literal("case"),
                            Reference('checkpoint_manager')
                        ),
                        body=List([Reference("id")])
                    )
                )
            )
        ])

        self._compare_minilinq_to_compiled(
            minilinq, '008_multiple-tables.xlsx', combine_emits=True
        )

    def test_multi_emit_no_combine(self):
        minilinq = List([
            Bind(
                "checkpoint_manager",
                Apply(
                    Reference('get_checkpoint_manager'), Literal("form"),
                    Literal(["Forms"])
                ),
                Emit(
                    table="Forms",
                    headings=[Literal("id"), Literal("name")],
                    missing_value='---',
                    source=Map(
                        source=Apply(
                            Reference("api_data"), Literal("form"),
                            Reference('checkpoint_manager')
                        ),
                        body=List([
                            Reference("id"),
                            Reference("form.name"),
                        ]),
                    )
                )
            ),
            Bind(
                "checkpoint_manager",
                Apply(
                    Reference('get_checkpoint_manager'), Literal("form"),
                    Literal(["Cases"])
                ),
                Emit(
                    table="Cases",
                    headings=[Literal("case_id")],
                    missing_value='---',
                    source=Map(
                        source=FlatMap(
                            body=Reference("form..case"),
                            source=Apply(
                                Reference("api_data"), Literal("form"),
                                Reference('checkpoint_manager')
                            )
                        ),
                        body=List([
                            Reference("@case_id"),
                        ]),
                    )
                )
            ),
            Bind(
                "checkpoint_manager",
                Apply(
                    Reference('get_checkpoint_manager'), Literal("case"),
                    Literal(["Other cases"])
                ),
                Emit(
                    table="Other cases",
                    headings=[Literal("id")],
                    missing_value='---',
                    source=Map(
                        source=Apply(
                            Reference("api_data"), Literal("case"),
                            Reference('checkpoint_manager')
                        ),
                        body=List([Reference("id")])
                    )
                )
            )
        ])

        self._compare_minilinq_to_compiled(
            minilinq, '008_multiple-tables.xlsx', combine_emits=False
        )

    def test_multi_emit_with_organization(self):
        minilinq = List([
            Bind(
                "checkpoint_manager",
                Apply(
                    Reference('get_checkpoint_manager'), Literal("form"),
                    Literal(["Forms", "Cases"])
                ),
                Filter(
                    predicate=Apply(Reference("filter_empty"), Reference("$")),
                    source=Map(
                        source=Apply(
                            Reference("api_data"), Literal("form"),
                            Reference('checkpoint_manager')
                        ),
                        body=List([
                            Emit(
                                table="Forms",
                                headings=[
                                    Literal("id"),
                                    Literal("name"),
                                    Literal("commcare_userid")
                                ],
                                missing_value='---',
                                source=Map(
                                    source=Reference("`this`"),
                                    body=List([
                                        Reference("id"),
                                        Reference("form.name"),
                                        Reference("$.metadata.userID"),
                                    ]),
                                )
                            ),
                            Emit(
                                table="Cases",
                                headings=[
                                    Literal("case_id"),
                                    Literal("commcare_userid")
                                ],
                                missing_value='---',
                                source=Map(
                                    source=Reference("form..case"),
                                    body=List([
                                        Reference("@case_id"),
                                        Reference("$.metadata.userID"),
                                    ]),
                                )
                            )
                        ])
                    )
                )
            ),
            Bind(
                'checkpoint_manager',
                Apply(
                    Reference('get_checkpoint_manager'), Literal("case"),
                    Literal(["Other cases"])
                ),
                Emit(
                    table="Other cases",
                    headings=[Literal("id"),
                              Literal("commcare_userid")],
                    missing_value='---',
                    source=Map(
                        source=Apply(
                            Reference("api_data"), Literal("case"),
                            Reference('checkpoint_manager')
                        ),
                        body=List([Reference("id"),
                                   Reference("$.user_id")])
                    )
                )
            )
        ])

        column_enforcer = ColumnEnforcer()
        self._compare_minilinq_to_compiled(
            minilinq,
            '008_multiple-tables.xlsx',
            combine_emits=True,
            column_enforcer=column_enforcer
        )

    def test_value_or_root(self):
        minilinq = List([
            Bind(
                "checkpoint_manager",
                Apply(
                    Reference('get_checkpoint_manager'), Literal("form"),
                    Literal(["Forms"])
                ),
                Emit(
                    table="Forms",
                    headings=[Literal("id"), Literal("name")],
                    missing_value='---',
                    source=Map(
                        source=Apply(
                            Reference("api_data"), Literal("form"),
                            Reference('checkpoint_manager')
                        ),
                        body=List([
                            Reference("id"),
                            Reference("form.name"),
                        ]),
                    )
                )
            ),
            Bind(
                "checkpoint_manager",
                Apply(
                    Reference('get_checkpoint_manager'), Literal("form"),
                    Literal(["Cases"])
                ),
                Emit(
                    table="Cases",
                    headings=[Literal("case_id")],
                    missing_value='---',
                    source=Map(
                        source=FlatMap(
                            body=Apply(
                                Reference("_or_raw"), Reference("form..case"),
                                Bind(
                                    "__root_only", Literal(True),
                                    Reference("$")
                                )
                            ),
                            source=Apply(
                                Reference("api_data"), Literal("form"),
                                Reference('checkpoint_manager')
                            )
                        ),
                        body=List([
                            Reference("@case_id"),
                        ]),
                    )
                )
            ),
            Bind(
                "checkpoint_manager",
                Apply(
                    Reference('get_checkpoint_manager'), Literal("case"),
                    Literal(["Other cases"])
                ),
                Emit(
                    table="Other cases",
                    headings=[Literal("id")],
                    missing_value='---',
                    source=Map(
                        source=Apply(
                            Reference("api_data"), Literal("case"),
                            Reference('checkpoint_manager')
                        ),
                        body=List([Reference("id")])
                    )
                )
            )
        ])

        self._compare_minilinq_to_compiled(
            minilinq,
            '008_multiple-tables.xlsx',
            combine_emits=False,
            value_or_root=True
        )

    def _compare_minilinq_to_compiled(self, minilinq, filename, **kwargs):
        print("Parsing {}".format(filename))
        abs_path = os.path.join(os.path.dirname(__file__), filename)
        compiled = get_queries_from_excel(
            openpyxl.load_workbook(abs_path), missing_value='---', **kwargs
        )
        assert compiled.to_jvalue() == minilinq.to_jvalue(), filename



"""rename_time_of_run_to_since_param

Revision ID: 29c27e7e2bf6
Revises: d3ce9dc9907a
Create Date: 2018-05-03 12:01:13.046003

"""
import sqlalchemy as sa
from alembic import op

revision = '29c27e7e2bf6'
down_revision = 'd3ce9dc9907a'
branch_labels = None
depends_on = None


def upgrade():
    meta = sa.MetaData(bind=op.get_bind())
    table = sa.Table('commcare_export_runs', meta, autoload=True)
    if 'since_param' not in table.c:
        op.alter_column(
            'commcare_export_runs', 'time_of_run',
            new_column_name='since_param',
            existing_type=sa.Unicode(32)
        )


def downgrade():
    op.alter_column('commcare_export_runs', 'since_param', new_column_name='time_of_run')



"""add args columns for project and commcare url

Revision ID: 53f1aad98e33
Revises: 9945abb4ec70
Create Date: 2018-08-14 12:39:55.543173

"""
from alembic import op
import sqlalchemy as sa


revision = '53f1aad98e33'
down_revision = '9945abb4ec70'
branch_labels = None
depends_on = None


def upgrade():
    url = op.get_bind().engine.url
    collation = 'utf8_bin' if 'mysql' in url.drivername else None
    op.add_column(
        'commcare_export_runs',
        sa.Column('project', sa.Unicode(32, collation=collation))
    )
    op.add_column(
        'commcare_export_runs',
        sa.Column('commcare', sa.Unicode(255, collation=collation))
    )

def downgrade():
    op.drop_column('commcare_export_runs', 'project')
    op.drop_column('commcare_export_runs', 'commcare')



import logging
from typing import Any, Dict
from typing import List as ListType
from typing import Optional

from commcare_export.env import Env
from commcare_export.misc import unwrap, unwrap_val
from commcare_export.repeatable_iterator import RepeatableIterator
from commcare_export.specs import TableSpec

logger = logging.getLogger(__name__)


class MiniLinq(object):
    """
    The abstract base class for MiniLinqs, and also the factory/registry
    for dispatching parsing, etc.
    """

    def eval(self, env: Env) -> Any:
        raise NotImplementedError()

    #### Factory methods ####

    _node_classes: Dict[str, 'MiniLinq'] = {}

    @classmethod
    def register(cls, clazz, slug=None):
        cls._node_classes[slug or clazz.__name__] = clazz

    @classmethod
    def from_jvalue(cls, jvalue):
        """
        The term `jvalue` is code for "the output of a JSON
        deserialization". This  module does not actually care about
        JSON, which is concrete syntax, but  only the corresponding data
        model of lists and string-indexed dictionaries.

        (since this data might never actually be a string, that layer is
        handled elsewhere)
        """

        # This is a bit wonky, but this method really should not be
        # inherited. So if we end up here from a subclass, it is broken.
        if not issubclass(MiniLinq, cls):
            raise NotImplementedError()

        if isinstance(jvalue, str):
            return jvalue

        elif isinstance(jvalue, list):
            # Leverage for literal lists of data in the code
            return [MiniLinq.from_jvalue(v) for v in jvalue]

        elif isinstance(jvalue, dict):
            # Dictionaries are reserved; they must always have exactly
            # one entry and it must be the AST node class
            if len(jvalue.values()) != 1:
                raise ValueError(
                    'JValue serialization of AST contains dict with number of slugs != 1'
                )
            slug = list(jvalue.keys())[0]

            if slug not in cls._node_classes:
                raise ValueError(
                    'JValue serialization of AST contains unknown node type: %s'
                    % slug
                )

            return cls._node_classes[slug].from_jvalue(jvalue)

    def to_jvalue(self):
        raise NotImplementedError()


class Reference(MiniLinq):
    """
    An MiniLinq referencing a datum or data. It is flexible about what
    the type of the environment is, but it must support using these as
    keys.
    """

    def __init__(self, ref):
        self.ref = ref  #parse_jsonpath(ref) #ref
        self.nested = isinstance(self.ref, MiniLinq)

    def eval(self, env):
        if self.nested:
            ref = self.ref.eval(env)
            return env.lookup(ref)
        return env.lookup(self.ref)

    def __eq__(self, other):
        return isinstance(other, Reference) and self.ref == other.ref

    @classmethod
    def from_jvalue(cls, jvalue):
        return cls(MiniLinq.from_jvalue(jvalue['Ref']))

    def to_jvalue(self):
        return {'Ref': self.ref.to_jvalue() if self.nested else self.ref}

    def __repr__(self):
        return '%s(%r)' % (self.__class__.__name__, self.ref)


class Literal(MiniLinq):
    """
    An MiniLinq wrapper around a python value. Returns exactly the value
    given to it. Note: when going to/from jvalue the contents are left
    alone, so it can be _used_ with a non-JSON encodable value, but
    cannot be encoded.
    """

    def __init__(self, v):
        self.v = v

    def eval(self, env):
        return self.v

    def __eq__(self, other):
        return isinstance(other, Literal) and self.v == other.v

    def __repr__(self):
        return '%s(%r)' % (self.__class__.__name__, self.v)

    @classmethod
    def from_jvalue(cls, jvalue):
        return cls(jvalue['Lit'])

    def to_jvalue(self):
        return {'Lit': self.v}


class Bind(MiniLinq):
    """
    Binds the results of an expression to a new name. Will be useful in
    writing exports by hand or debugging, and maybe for efficiency if it
    de-dupes computation (but generally exports will be expected to be
    too large to store, so it'll be re-run on each access.
    """

    def __init__(self, name: str, value: MiniLinq, body: MiniLinq) -> None:
        self.name = name
        self.value = value
        self.body = body

    def eval(self, env):
        return self.body.eval(env.bind(self.name, self.value.eval(env)))

    def __eq__(self, other):
        return isinstance(
            other, Bind
        ) and self.name == other.name and self.value == other.value and self.body == other.body

    def __repr__(self):
        return '%s(name=%r, value=%r, body=%r)' % (
            self.__class__.__name__, self.name, self.value, self.body
        )

    @classmethod
    def from_jvalue(cls, jvalue):
        fields = jvalue['Bind']
        return cls(
            name=fields['name'],
            value=MiniLinq.from_jvalue(fields['value']),
            body=MiniLinq.from_jvalue(fields['body'])
        )

    def to_jvalue(self):
        return {
            'Bind': {
                'name': self.name,
                'value': self.value.to_jvalue(),
                'body': self.body.to_jvalue()
            }
        }


class Filter(MiniLinq):
    """
    Just what it sounds like
    """

    def __init__(
        self,
        source: MiniLinq,
        predicate: MiniLinq,
        name: Optional[str] = None
    ) -> None:
        self.source = source
        self.name = name
        self.predicate = predicate

    def eval(self, env):
        source_result = self.source.eval(env)

        # Python closure workaround
        def iterate(
            env_=env,
            source_result_=source_result,
        ):
            if self.name:
                for item in source_result_:
                    if self.predicate.eval(env_.bind(self.name, item)):
                        yield item
            else:
                for item in source_result_:
                    if self.predicate.eval(env_.replace(item)):
                        yield item

        return RepeatableIterator(iterate)

    def __eq__(self, other):
        return (
            isinstance(other, Filter) and self.source == other.source
            and self.name == other.name and self.predicate == other.predicate
        )

    @classmethod
    def from_jvalue(cls, jvalue):
        fields = jvalue['Filter']

        # TODO: catch errors and give informative error messages
        return cls(
            predicate=MiniLinq.from_jvalue(fields['predicate']),
            source=MiniLinq.from_jvalue(fields['source']),
            name=fields.get('name')
        )

    def to_jvalue(self):
        return {
            'Filter': {
                'predicate': self.predicate.to_jvalue(),
                'source': self.source.to_jvalue(),
                'name': self.name
            }
        }

    def __repr__(self):
        return '%s(source=%r, name=%r, predicate=%r)' % (
            self.__class__.__name__, self.source, self.name, self.predicate
        )


class List(MiniLinq):
    """
    A list of expressions, embeds the [ ... ] syntax into the MiniLinq
    meta-leval
    """

    def __init__(self, items):
        self.items = items

    def eval(self, env):
        return [item.eval(env) for item in self.items]

    def __eq__(self, other):
        return isinstance(other, List) and self.items == other.items

    def __repr__(self):
        return '%s(%s)' % (self.__class__.__name__, self.items)

    @classmethod
    def from_jvalue(cls, jvalue):
        return cls([MiniLinq.from_jvalue(item) for item in jvalue['List']])

    def to_jvalue(self):
        return {'List': [item.to_jvalue() for item in self.items]}


class Map(MiniLinq):
    """
    Like the `FROM` clause of a SQL `SELECT` or jQuery's map, binds each
    item from its `source` and evaluates the body MiniLinq.

    If `name` is provided to the constructor, then instead of replacing
    the environment with each row, it will just bind the row to `name`,
    enabling references to the rest of the env.
    """

    def __init__(
        self,
        source: MiniLinq,
        body: MiniLinq,
        name: Optional[str] = None
    ) -> None:
        self.source = source
        self.name = name
        self.body = body

    def eval(self, env):
        source_result = self.source.eval(env)

        def iterate(
            env=env, source_result=source_result
        ):  # Python closure workaround
            if self.name:
                for item in source_result:
                    yield self.body.eval(env.bind(self.name, item))
            else:
                for item in source_result:
                    yield self.body.eval(env.replace(item))

        return RepeatableIterator(iterate)

    def __eq__(self, other):
        return (
            isinstance(other, Map) and self.name == other.name
            and self.source == other.source and self.body == other.body
        )

    @classmethod
    def from_jvalue(cls, jvalue):
        fields = jvalue['Map']

        # TODO: catch errors and give informative error messages
        return cls(
            body=MiniLinq.from_jvalue(fields['body']),
            source=MiniLinq.from_jvalue(fields['source']),
            name=fields.get('name')
        )

    def to_jvalue(self):
        return {
            'Map': {
                'body': self.body.to_jvalue(),
                'source': self.source.to_jvalue(),
                'name': self.name
            }
        }


class FlatMap(MiniLinq):
    """
    Somewhat like a JOIN, but not quite. Called `SelectMany` in LINQ and
    `flatMap` other languages. Obvious equivalence: `flatMap f = flatten
    . map f` but so common it is useful to have around.

    If `name` is provided to the constructor, then instead of replacing
    the environment with each row, it will just bind the row to `name`,
    enabling references to the rest of the env.
    """

    def __init__(
        self,
        source: MiniLinq,
        body: MiniLinq,
        name: Optional[str] = None
    ) -> None:
        self.source = source
        self.name = name
        self.body = body

    def eval(self, env):
        source_result = self.source.eval(env)

        # Python closure workaround
        def iterate(
            env_=env,
            source_result_=source_result,
        ):
            if self.name:
                for item in source_result_:
                    for result_item in self.body.eval(
                        env_.bind(self.name, item)
                    ):
                        yield result_item
            else:
                for item in source_result_:
                    for result_item in self.body.eval(env_.replace(item)):
                        yield result_item

        return RepeatableIterator(iterate)

    def __eq__(self, other):
        return (
            isinstance(other, FlatMap) and self.name == other.name
            and self.source == other.source and self.body == other.body
        )

    @classmethod
    def from_jvalue(cls, jvalue):
        fields = jvalue['FlatMap']

        # TODO: catch errors and give informative error messages
        return cls(
            body=MiniLinq.from_jvalue(fields['body']),
            source=MiniLinq.from_jvalue(fields['source']),
            name=fields.get('name')
        )

    def to_jvalue(self):
        return {
            'FlatMap': {
                'body': self.body.to_jvalue(),
                'source': self.source.to_jvalue(),
                'name': self.name
            }
        }


class Apply(MiniLinq):
    """
    Abstract syntax for function or operator application.
    """

    def __init__(self, fn, *args):
        self.fn = fn
        self.args = args

    def eval(self, env):
        fn_result = self.fn.eval(env)
        args_results = [arg.eval(env) for arg in self.args]

        try:
            result = fn_result(*args_results)
            if isinstance(result, MiniLinq):
                return result.eval(env)
        except Exception as e:
            args = ', '.join([str(unwrap_val(arg)) for arg in args_results])
            try:
                doc_id = unwrap_val(Reference('id').eval(env))
            except:
                doc_id = 'unknown'

            message = e.args[0] + (
                f": Error processing document '{doc_id}'. Failure to "
                f"evaluating expression '{self!r}' with arguments '{args}'"
            )

            e.args = (message,) + e.args[1:]
            raise
        return result

    def __eq__(self, other):
        return (
            isinstance(other, Apply) and self.fn == other.fn
            and self.args == other.args
        )

    @classmethod
    def from_jvalue(cls, jvalue):
        fields = jvalue['Apply']

        # TODO: catch errors and give informative error messages
        return cls(
            MiniLinq.from_jvalue(fields['fn']),
            *[MiniLinq.from_jvalue(arg) for arg in fields['args']]
        )

    def to_jvalue(self):
        return {
            'Apply': {
                'fn': self.fn.to_jvalue(),
                'args': [arg.to_jvalue() for arg in self.args]
            }
        }

    def __repr__(self):
        return f'{self.__class__.__name__}({self.fn!r}, *{self.args!r})'


class Emit(MiniLinq):
    """
    This MiniLinq writes a whole table to whatever writer is registered
    in the `env`. In practice,  a table to a dict of a name, headers,
    and rows, so the writer is free to do an idempotent upsert, etc.

    Note that it does not actually check that the number of headings is
    correct, nor does it try to ensure that the things being emitted are
    actually lists - it is just crashy instead.
    """

    def __init__(
        self,
        table: str,
        headings: ListType[MiniLinq],
        source: MiniLinq,
        missing_value: Optional[str] = None,
        data_types: Optional[ListType[Literal]] = None,
    ) -> None:
        """
        Initializes an ``Emit`` instance.

        :param table: The name/title of the table to be written.
        :param headings: Evaluated to determine column headings.
        :param source: Evaluated to determine the table rows.
        :param missing_value: Denotes "no value". e.g. ``"---"``
        :param data_types: The data types of the columns. e.g.
            ``[Literal('text'), Literal('date'), ...]``
        """
        self.table = table
        self.headings = headings
        self.source = source
        self.missing_value = missing_value
        self.data_types = data_types or []

    @unwrap('cell')
    def coerce_cell_blithely(self, cell):
        if isinstance(cell, list):
            if not cell:  # jsonpath returns empty list when path is not present
                return self.missing_value
            return ','.join([self.coerce_cell(item) for item in cell])
        else:
            return cell

    def coerce_cell(self, cell):
        try:
            return self.coerce_cell_blithely(cell)
        except Exception:
            logger.exception(
                'Error converting value to exportable form: %r' % cell
            )
            return ''

    def coerce_row(self, row):
        return [self.coerce_cell(cell) for cell in row]

    def eval(self, env):
        rows = self.source.eval(env)
        env.emit_table(
            TableSpec(
                name=self.table,
                headings=[heading.eval(env) for heading in self.headings],
                rows=map(self.coerce_row, rows),
                data_types=[lit.v for lit in self.data_types]
            )
        )

    @classmethod
    def from_jvalue(cls, jvalue):
        fields = jvalue['Emit']
        return cls(
            table=fields['table'],
            source=MiniLinq.from_jvalue(fields['source']),
            headings=[
                MiniLinq.from_jvalue(heading) for heading in fields['headings']
            ],
            missing_value=fields.get('missing_value'),
            data_types=fields.get('data_types'),
        )

    def to_jvalue(self):
        return {
            'Emit': {
                'table':
                    self.table,
                'headings': [heading.to_jvalue() for heading in self.headings],
                'source':
                    self.source.to_jvalue(),
                'missing_value':
                    self.missing_value,
                'data_types': [
                    heading.to_jvalue() for heading in self.headings
                ]
            }
        }

    def __eq__(self, other):
        return (
            isinstance(other, Emit) and self.table == other.table
            and self.headings == other.headings and self.source == other.source
            and self.missing_value == other.missing_value
            and self.data_types == other.data_types
        )

    def __repr__(self):
        return '%s(table=%r, headings=%r, source=%r, missing_value=%r)' % (
            self.__class__.__name__,
            self.table,
            self.headings,
            self.source,
            self.missing_value
        )


### Register everything with the root parser ###

MiniLinq.register(Reference, slug='Ref')
MiniLinq.register(Literal, slug='Lit')
MiniLinq.register(Map)
MiniLinq.register(Filter)
MiniLinq.register(FlatMap)
MiniLinq.register(Apply)
MiniLinq.register(Emit)
MiniLinq.register(List)
MiniLinq.register(Bind)
